
;;  Paradigmata Programovani 2A, PRIKLADY
;;  Copyright (C) 2008  Vilem Vychodil, <vilem.vychodil@upol.cz>
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; version 2 of the License.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;; PREDNASKA 1: Imperativni rysy, stav, prikaz prirazeni

;; 1. CAST: SEKVENCOVANI VYRAZU A VEDLEJSI EFEKT
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; UVOD

;; funkcionalni programovani = zalozene na ,,vraceni hodnot''
;; imperativni (proceduralni) programovani = zalozene na provadeni
;;                                           vedlejsich efektu

;; vedlejsi efekt: souvisi (typicky) se
;; a) vstupem / vystupem (nacitani, ukladani dat)
;; b) zmenou vazby symbolu (set!, uvidime dnes)
;; c) destruktivni zmenou datove struktury (set-car! a spol., priste)

;;; SEKVENCOVANI: forma begin a procedura for-each
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; begin: postupne se v aktualnim prostredi vyhodnoti vsechny argumenty,
;;        vraci se hodnota posledniho z nich
;;        hodnoty predchozich argumentu jsou ,,zapomenuty''
;;        predpoklada se ze vyhodnocovani techto argumentu bude mit
;;        vedlejsi efekt (side effect)

;; specialni forma begin -- sekvencovani
;; hodnota 3 vznikla vyhodnocenim (+ 1 2) bude zapomenuta
(begin
  (+ 1 2)
  (* 3 4))

;; zde uz mame vedlejsi efekt -- vytisteni na obrazovku
(begin
  (display (+ 1 2))
  (newline)
  (* 3 4))

;; POZOR: (display "Ahoj") vypise Ahoj na obrazovku -- to je side efekt,
;;        nejedna se o hodnotu vyhodnoceni display (display vraci nedef.
;;        hodnotu, ostatne skoro jako vsechny zabudovane procedury provadejici
;;        side efekt

(begin
  (display "Vysledek je ")
  (display (* 3 4))
  (newline))

;; vyhodnocovani vyrazu probiha v aktualnim prostredi
(define x 100)

(begin
  (define x 10)
  (* 2 x))

x	; nova hodnota bude 10

;; tohle je begin v novem prostredi
((lambda ()
   (begin
     (define x 10)
     (* 2 x))))

;; prepis predchoziho
(let ()
  (begin
    (define x 10)
    (* 2 x)))

;; pozn.: predchozi lze zkratit (vyjmutim nadbytecneho ``begin'')

;; begin lze pouzit i bez tela
(begin)		; nedefinovana hodnota

;; K CEMU JE begin + display DOBRY?
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; napriklad pro odladovani:
(define depth-map
  (lambda (f l)
    (cond ((null? l) '())
	  ((not (list? (car l))) (cons (f (car l)) (depth-map f (cdr l))))
	  (else (cons (depth-map f (car l))
		      (depth-map f (cdr l)))))))

(depth-map - '(1 (2 ((3)) 4) 5))

;; na tomto budeme videt, jak probiha vypocetni proces
(define depth-map
  (lambda (f l)
    (display (list "CALL:" l))
    (newline)
    (cond ((null? l)
	   (begin
	     (display "PRAZDNY")
	     (newline)
	     '()))
	  ((not (list? (car l)))
	   (begin
	     (display "ZPRACUJ ELEMENT ")
	     (display (car l))
	     (display " -> ")
	     (display (f (car l)))
	     (newline)
	     (cons (f (car l))
		   (depth-map f (cdr l)))))
	  (else
	   (begin
	     (display "ZPRACUJ SEZNAMY ")
	     (display (car l))
	     (display " a ")
	     (display (cdr l))
	     (newline)
	     (cons (depth-map f (car l))
		   (depth-map f (cdr l))))))))

(depth-map - '(1 (2 ((3)) 4) 5))

;; proved prikazy pro kazdy prvek seznamu
(for-each
 (lambda (x)
   (display "Cislo: ")
   (display x)
   (newline))
 '(10 20 30 40))	; vysledkem je nedef. hodnota (vzdy)

; rozdil oproti map je videt napriklad na tomto:
(for-each (lambda (x) x) '(1 2 3))	; nedef.
(map (lambda (x) x) '(1 2 3))		; (1 2 3)

;; muze se pouzit i s vice seznamy, analogicky jako map:
(for-each
 (lambda (x y z)
   (newline)
   (display (list x y z))
   (newline))
 '(10 20) '(#t #f) '(a b))

;; for-each lze jednoduse naprogramovat (verze pro jeden seznam):
;; vzimnete si pouziti `if' bez alternativniho vyrazu
(define for-each
  (lambda (f l)
    (if (not (null? l))
	(begin
	  (f (car l))
	  (for-each f (cdr l))))))

;; a obecna verze je taky jednoducha:
(define for-each
  (lambda (f . lists)
    (if (not (null? (car lists)))
	(begin
	  (apply f (map car lists))
	  (apply for-each f (map cdr lists))))))

;;; PRIKAZ PRIRAZENI: set!
;;;;;;;;;;;;;;;;;;;;;;;;;;

;; set! je specialni forma (syntaxe stejna jako u define)
;; rict jak funguje: postupne prochazi prostredi, zacina se lokalnim a 
;; postupuje se pres predky az ke globalnimu, pritom se vzdy v aktualnim
;; prostredi kontroluje, zda-li ma dany symbol vazbu, pokud ano, vazba
;; se nahradi hodnotou vzniklou vyhodnocenim vyrazu

;; srovnej:
(define x 100)

(let ()
  (define x 10)
  (+ x 1))

x

;; versus
(define x 100)

(let ()
  (set! x 10)
  (+ x 1))

x

;; poznamka: kdysi jsme rozsirili lambda -- v tele bylo vic vyrazu,
;; nyni to zase muzeme ,,vzit zpet'':

;; pri aktivaci procedury
(lambda (x)
  (display "Blah")
  (+ x 1))

;; je vyhodnoceno nasledujici telo (v lokalnim prostredi)
(begin
  (display "Blah")
  (+ x 1))

;; 2. CAST: IMPERATIVNI RYSY V PROGRAMOVANI
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; funkcionalni program: posloupnost symbolickych vyrazu
;;              vypocet: postupne vyhodnocovani vyrazu
;;                       dochazi pri nem k aplikaci procedur

;; proceduralni program: nestrukturovany/strukturovany kod obsahujici
;;                       prikazy prirazeni, vetveni a skoku
;;              vypocet: sekvencni provadeni prikazu, ktere maji zpravidla
;;                       vedlejsi efekt (zejmena prirazeni), vetveni vypoctu je
;;                       na zaklade podminenych/nepodminenych skoku (navesti)

;; prikaz prirazeni: LVALUE := RVALUE (misto := se pouzivaji i jine, C ma =)
;; LVALUE ... pametove misto (adresa)
;; RVALUE ... hodnota
;; pri provedeni prikazu je na pametove misto LVALUE ulozena hodnota RVALUE

;; funkcionalni jazyky maji (zpravidla) proceduralni rysy
;; napr. Scheme, Common LISP maji cykly, ruzne typy skoku, prikazy prirazeni

;; proceduralni jazyky zridka kdy maji funkcionalni rysy
;; napr. v C, C++, Pascal, Java se neda s proc. pracovat jako s elem. 1 radu
;; jezna z vyjimek: PERL ma lambda-vyrazy, lexikalni uzavery, zabudovany map

;;; PROGRAMOVANI S PRIKAZEM PRIRAZENI
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; iterativni vypocet fibonacciho cisel: ,,funkcionalne ciste reseni''
(define fib
  (lambda (n)
    (let iter ((a 1)
	       (b 1)
	       (i n))
      (if (<= i 1)
	  a
	  (iter b (+ a b) (- i 1))))))

;; iterativni vypocet fibonacciho cisel: ,,imperativne ciste reseni''
(define fib
  (lambda (n)
    (define a 1)
    (define b 1)
    (define c 'pomocna-promenna)	; zavadime kvuli prohozeni hodnot 
    (define i n)
    ;; nasledujici procedurou bez argumentu vytvorime ,,cyklus''
    (define loop
      (lambda ()
	(if (> i 1)
	    (begin
	      (set! c b)
	      (set! b (+ a b))
	      (set! a c)
	      (set! i (- i 1))
	      (loop)))))
    ;; spust cyklus
    (loop)
    ;; vrat hodnotu ulozenou v ``a''
    a))

;; ZAVER: Scheme dokaze byt ,,plnohodnotny imperativni (proceduralni) jazyk''
;;        ale je potreba pouzivat s mirou

;; nekdy se pridani imperativniho rysu do programu hodi
;; zde napriklad pocitame indexy prvku v seznamu
(let ((i 0))	; nastav i na 0
  (for-each
   (lambda (x)
     (display "Index: ")
     (display i)
     (display " , prvek: ")
     (display x)
     (newline)
     (set! i (+ i 1)))	; zvedni i o 1
   '(a b c d e f)))

;;; PROCEDURY DRZICI SI VLASTNI STAV
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; globalne zavedeny symbol
(define value 0)

;; procedura pracujici s ,,globalnim symbolem''
(define inc
  (lambda (x)
      (set! value (+ value x))
      value))

(inc 10)
(inc 10)
(inc 10)

value

;; docasna zmena hodnoty lexikalni promenne
;; specialni forma fluid-let (nestandardni, neni v R5RS)

;; syntax je stejna jako u formy `let', nevytvari se ale nove prostredi,
;; pouze se zmeni vazby symbolu a po ukonceni fluid-let se zmeni
;; na puvodni hodnoty
(fluid-let ((value 200))
  (display (inc 10))
  (newline)
  (display (inc 10))
  (newline))

(inc 10)

;; srovnej s:
(let ((value 200))
  (display (inc 10))
  (newline)
  (display (inc 10))
  (newline))

;; procedura udrzujici si vlastni vnitrni stav
;; misto globalniho symbolu menime vazbu symbolu na ktery neni videt ,,zvenci''
(define inc
  (let ((value 0))
    (lambda (x)
      (set! value (+ value x))
      value)))

(inc 5)
(inc 5)
(inc 5)

value

;; POZOR: nasledujici nefunguje (zamena `lambda' a `let')
(define inc
  (lambda (x)
    (let ((value 0))
      (set! value (+ value x))
      value)))

;; HROZBA: pri pouzivani procedur s vedlejsim efektem se podstatne hur ladi
;;         uz nestaci pouze testovat vstupy a vystupy procedury, musime
;;         vzdy pocitat s tim, jaky ma procedura zrovna vnitrni stav!

;; PRIKLAD

;; pricita jednicku (predavane argumenty jsou uplne ignorovany)
(define inc1
  (let ((value 0))
    (lambda args
      (set! value (+ value 1))
      value)))

;; build-list pracujici ,,zepredu''
(define build-list
  (lambda (n f)
    (let iter ((i 0))
      (if (= i n)
	  '()
	  (cons (f i) (iter (+ i 1)))))))

(build-list 10 (lambda (x) x))	; (0 1 2 3 4 5 6 7 8 9)

;; iterativni build-list pracujici ,,zezadu''
(define build-list
  (lambda (n f)
    (let iter ((i (- n 1))
	       (aux '()))
      (if (< i 0)
	  aux
	  (iter (- i 1) (cons (f i) aux))))))

; od predchoziho je z funkcionalniho pohledu nerozlisitelny
(build-list 10 (lambda (x) x))	; (0 1 2 3 4 5 6 7 8 9)

; muzeme je ale odlisit pomoci vedlejsiho efektu
(build-list 10 inc1)
; v prvnim pripade: (1 2 3 4 5 6 7 8 9 10)
; v druhem pripade: (10 9 8 7 6 5 4 3 2 1)

; pokud spustime v prvnim pripade 3x po sobe, obdrzime 3 ruzne vysledky
(build-list 10 inc1)
; (1 2 3 4 5 6 7 8 9 10)
; (11 12 13 14 15 16 17 18 19 20)
; (21 22 23 24 25 26 27 28 29 30)
; ...

;; VSIMNETE SI: zde je videt, ze pri pouziti procedur s vedlejsim efektem
;;              se obecne muzeme dostat pri ladeni do komplikovane situace:
;;              procedura se pri nasobnem spusteni ,,chova jinak''

;; PRIKLAD: zmena platnosti podminky pri druhem vyhodnoceni

; vlastnost
(define prop?
  (lambda (x)
    (< x 3)))

; test vlastnosti prvku
(let ((x 2))
  (if (prop? x) (prop? x) 'blah))

; predikat pracujici s vnitrnim stavem
(define prop?
  (let ((y 'whatever)
	(z 0))
    (lambda (x)
      (set! y (< z 3))
      (set! z x)
      y)))

; nyni je podminka pri prvnim zavolani pravdiva, pri druhem jiz ne!
(let ((x 3))
  (if (prop? x) (prop? x) 'blah))

;;; PROCEDURY GENERUJICI PROCEDURY S VNITRNIM STAVEM
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; MOTIVACE: jak to provest, pokud budeme chtit mit ,,vic nezavislych scitacu''

;; procedura (bez argumentu), ktera vraci proceduru jednoho argumentu
;; reprezentujici jeden scitac (instance)
(define make-inc
  (lambda ()
    (let ((value 0))
      (lambda (x)
	(set! value (+ value x))
	value))))

(define i (make-inc))
(define j (make-inc))
(define k (make-inc))

i	; #<procedure>
j	; #<procedure>
k	; #<procedure>

(i 1)
(i 1)
(i 1)
(j 5)
(j 5)
(j 5)
(j 0)
(k 10)
(k 20)
(k 0)

;; novy map, ktery spolu s prvkem predava i jeho index
;; vyresili jsme pomoci map a procedury, ktera vyuziva zmenu stavu
(define map-index
  (lambda (f l)
    (let ((i -1))
      (map (lambda (x)
	     (set! i (+ i 1))
	     (f i x))
	   l))))

(map-index cons '(a b c d e))	; ((0 . a) (1 . b) (2 . c) (3 . d) (4 . e))

;; v predchozi ukazce jsme tise predpokladali, ze map funguje ,,odpredu''
;; pokud nekdo zmeni samotny map, dostaneme se do maleru:

;; ukazkova verze map, ktera pracuje ,,odzadu''
(define map
  (lambda (f l)
    (let iter ((l (reverse l))
	       (aux '()))
      (if (null? l)
	  aux
	  (iter (cdr l) (cons (f (car l)) aux))))))

(map-index cons '(a b c d e))	; ((4 . a) (3 . b) (2 . c) (1 . d) (0 . e))

;;; PROCEDURY SDILEJICI TYZ STAV
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; procedury s vedlejsim efektem lze pouzit ke sniffovani jinych procedur
;; muzeme ,,nahlednout'' do toho, co delaji jine procedury

;; Priklad: muzeme zjistit, jak dany tridici algoritmus postupuje pri
;; porovnavani prvku, aniz bychom museli upravit zdrojak tridiciho algoritmu
;; (muzeme tak odhadnout o jaky algoritmus se jedna)

;; vytvoreni ,,chytreho predikatu'',
;; ktery si pamatuje s jakymi hodnotami byl volan
(define make-pred
  (lambda (pred?)
    (let ((called-with-values '()))
      `(,(lambda (x y)
	   (begin
	     (set! called-with-values
		   (cons (cons x y) called-with-values))
	   (pred? x y)))
	,(lambda ()
	   (reverse called-with-values))))))

;; nyni muzeme videt, jak mergesort postupne porovnava prvky
(let* ((pred (make-pred <=))
       (new-pred? (car pred))
       (get-called (cadr pred)))
  (display (mergesort '(4 2 8 5 6 4 5 3 5 6 7 8) new-pred?))
  (newline)
  (get-called))

;; nebo quicksort
(let* ((pred (make-pred <=))
       (new-pred? (car pred))
       (get-called (cadr pred)))
  (display (quicksort '(4 2 8 5 6 4 5 3 5 6 7 8) new-pred?))
  (newline)
  (get-called))

;; 3. CAST: JEDNODUCHY OBJEKTOVY SYSTEM: INSTANCIOVANI, ZASILANI SIGNALU
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; OBJEKTOVE PARADIGMA: primo se jim nebudeme zabyvat (PP3), zakladni myslenka,
;; programator spravuje system elementu (objektu), ktere maji svuj stav
;; vypocetni proces je potom slozen ze vzajemne interakce objektu a zmen
;; jejich stavu (nejznamejsi obektove systemy jsou zalozeny na zasilani (emisi)
;; signalu a jejich prijmu (pomoci slotu), tzv. message-passing style of
;; programming)

;; nektere problemy se takto resi pohodlne -- GUI .. okna maji svuj stav
;; (pozice, focus, mohou obsahovat dalsi okna, etc.)

;; jednoduchy objektovy system -- instanciovani, zasilani signalu

;; modul zasobnik (bez instancovani)
(define stack

  ;; data (vnitrni stav objektu)
  (let ((stack '())
	(depth 0))

    ;; je zasobnik prazdny?
    (define is-empty?
      (lambda ()
	(= depth 0)))

    ;; pridej prvek na vrchol
    (define push
      (lambda (elem)
	(set! stack (cons elem stack))
	(set! depth (+ depth 1))))

    ;; odstran prvek z vrcholu
    (define pop
      (lambda ()
	(if (is-empty?)
	    (error "cannot perform `pop', stack is empty")
	    (begin
	      (set! stack (cdr stack))
	      (set! depth (- depth 1))))))

    ;; vrat prvek nachazejici se na vrcholu
    (define top
      (lambda ()
	(if (is-empty?)
	    (error "stack is empty")
	    (car stack))))

    ;; dispatch (aktivuje jednotlive procedury podle jmen signalu)
    (lambda (signal . args)
      (cond ((equal? signal 'empty?) (is-empty?))
	    ((equal? signal 'push) (push (car args)))
	    ((equal? signal 'pop) (pop))
	    ((equal? signal 'top) (top))
	    (else (error "unknown signal"))))))

; priklad pouziti
(stack 'push 10)
(stack 'top)
(stack 'push 20)
(stack 'top)
(stack 'push 30)
(stack 'top)
(stack 'empty?)
(stack 'pop)
(stack 'pop)
(stack 'pop)
(stack 'empty?)
(stack 'pop)

;; zasobnik, ktery se instanciuje
(define make-stack

  (lambda ()

    ;; data
    (let ((stack '())
	  (depth 0))

      ;; je zasobnik prazdny?
      (define is-empty?
	(lambda ()
	  (= depth 0)))

      ;; pridej prvek na vrchol
      (define push
	(lambda (elem)
	  (set! stack (cons elem stack))
	  (set! depth (+ depth 1))))

      ;; odstran prvek z vrcholu
      (define pop
	(lambda ()
	  (if (is-empty?)
	      (error "cannot perform `pop', stack is empty")
	      (begin
		(set! stack (cdr stack))
		(set! depth (- depth 1))))))

      ;; vrat prvek nachazejici se na vrcholu
      (define top
	(lambda ()
	  (if (is-empty?)
	      (error "stack is empty")
	      (car stack))))

      ;; dispatch (aktivuje jednotlive procedury podle jmen signalu)
      (lambda (signal . args)
	(cond ((equal? signal 'empty?) (is-empty?))
	      ((equal? signal 'push) (push (car args)))
	      ((equal? signal 'pop) (pop))
	      ((equal? signal 'top) (top))
	      (else (error "unknown signal")))))))

; priklad pouziti
(define s1 (make-stack))
(define s2 (make-stack))
(s1 'push 10)
(s2 'push 'a)
(s1 'push 20)
(s2 'push 'b)
(s1 'push 30)
(s1 'top)
(s2 'top)
