
;;  Paradigmata Programovani 2A, PRIKLADY
;;  Copyright (C) 2008  Vilem Vychodil, <vilem.vychodil@upol.cz>
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; version 2 of the License.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; MUTOVATELNE PARY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; ZAKLADY PRACE
;;;;;;;;;;;;;;;;;

;; nasledujicim nedojde ke zmene prvku, ale rovnou k vytvoreni noveho seznamu
(define s '(1 2 3))
(set! s '(1 blah 3))

;; mutatory paru: procedury set-car!, set-cdr!
(define p (cons 1 2))
p
(set-car! p 'ahoj)
p
(set-cdr! p 'svete)
p

;; Priklad:
(define s '(1 2 3))
(set-car! (cdr s) 'blah)
s

;; provazene seznamy, nutno dbat zvysene obezretnosti!
;; riziko vzniku chyb, tak ze si ,,nechtene zmenime neco jineho''
(define s '(1 2 3))
(define r (list 10 s 20))
r
(set-car! (cadr r) 'neco)
r
s

;; Priklad:

;; pro n=4 vytvor seznam tvaru: ((#f #f #f #f) (#f #f #f) (#f #f) (#f)), etc.
;; stupidni (primocare) reseni (malo efektivni)
;; slozitost: O(n(1+n)/2)
(define f-list
  (lambda (n)
    (build-list n
		(lambda (i)
		  (build-list (- n i)
			      (lambda (x) #f))))))

; nasledujici asi neprekvapi:
(define s (f-list 4))
s
(set-car! (car (reverse s)) 'blah)
s
(set-car! (cadr (reverse s)) 100)
s	; ((#f #f #f #f) (#f #f #f) (100 #f) (blah))

;; ta sama procedura ale efektivnejsi, vyuziva faktu, ze pokud mame
;; vytvorenou kolekci pro n, pak kolekci pro n+1 muzeme ziskat v konst. case
;; slozitost: O(n)
(define f-list
  (lambda (n)
    (if (= n 1)
	'((#f))
	(let ((rest (f-list (- n 1))))
	  (cons (cons (caar rest) (car rest)) rest)))))

; nasledujici mozna prekvapi:
(define s (f-list 4))
s
(set-car! (car (reverse s)) 'blah)
s
(set-car! (cadr (reverse s)) 100)
s	; ((#f #f 100 blah) (#f 100 blah) (100 blah) (blah))

;; VELKY PROBLEM: mutace je nebezpecna: bez znalosti toho, jak je dana
;; procedura naprogramovana jiz nemuzeme vytvaret ,,bezpecny kod'' .. kod
;; o kterem muzeme dokazat, ze vykonava to, co chceme!

;; To je v primem rozporu s konceptem procedury jako CERNE SKRINKY (BLACK BOX)!

;; jelikoz program = data, data = program, pomoci set-car! a set-cdr!
;; je de facto mozne destruktivne modifikovat samotny program (za behu):

;; Priklad:

(define proc 
  (lambda (x)
    (display (list "Input parameter: " x))
    (newline)
    (set-car! x (+ (car x) 1))
    x))

(define test
  (lambda ()
    (proc '(0))))

(test)
(test)

;; seznam (0), ktery je v procedure test predavan procedure proc je pri
;; jeji prvni aplikaci destruktivne zmenen. Pri druhe aplikaci je jiz zmeneny

;; DVA POSTREHY:
;; 1) zmutovali jsme de facto ,,telo procedury'', konkretne jsme zmutovali
;;    jednoprvkovy seznam (0), ktery je predavan pri volani procedure proc
;; 2) telo uzivatelske procedury test, to jest seznam (proc '(0)) je po
;;    prvnim volani zmeneno na (proc '(1)), po druhem volani na (proc '(1)),
;;    uvedomte si, ze procedura proc je aplikovana se seznamem (0), (1), ...
;;    ktery je vzdy na stejnem miste v pameti -- specialni forma quote pouze
;;    vraci svuj argument -- coz je porad tyz seznam, ktery byl vytvoren
;;    pri prevedeni symbolickeho vyrazu (0) do jeho interni formy
;; 3) z prikladu je videt, ze data a program jsou skutecne totez, kdyby
;;    nebyly, vyse uvedena vec by nesla provest (pouzili jsme mutatory
;;    na data k mutaci programu, tj. program se musi chovat jako data)
;; 4) na 5. prednasce si ukazeme daleko vic :-)

;;; VYTVORENI MUTOVATELNYCH PARU POMOCI PROCEDUR VYSSICH RADU
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstrukce mutovatelneho paru
(define cons
  (lambda (x y)

    ;; modifikator vazby symbolu x
    (define set-x!
      (lambda (value)
	(set! x value)))

    ;; modifikator vazby symbolu y
    (define set-y!
      (lambda (value)
	(set! y value)))

    ;; dispatch
    (lambda (signal)
      (cond ((equal? signal 'car) x)
	    ((equal? signal 'cdr) y)
	    ((equal? signal 'set-car!) set-x!)
	    ((equal? signal 'set-cdr!) set-y!)
	    (else 'unknown-signal)))))

;; selektor car
(define car
  (lambda (pair)
    (pair 'car)))

;; selektor cdr
(define cdr
  (lambda (pair)
    (pair 'cdr)))

;; mutace prvniho prvku
(define set-car!
  (lambda (pair value)
    ((pair 'set-car!) value)))

;; mutace druheho prvku
(define set-cdr!
  (lambda (pair value)
    ((pair 'set-cdr!) value)))

(define s (cons 10 (cons 20 (cons 30 '()))))
(car s)
(car (cdr s))
(car (cdr (cdr s)))
(set-car! (cdr s) 100)
(car s)
(car (cdr s))
(car (cdr (cdr s)))
(set-cdr! (cdr s) 200)
(car s)
(car (cdr s))
(cdr (cdr s))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; DESTRUKTIVNI PRACE SE SEZNAMY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; DESTRUKTIVNI MODIFIKACE PRVKU SEZNAMU
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; klasicka nedestruktivni (funkcionalni) verze
(define list-set
  (lambda (l index value)
    (let iter ((l l)
	       (i 0))
      (if (= i index)
	  (cons value (cdr l))
	  (cons (car l) (iter (cdr l) (+ i 1)))))))

(define s '(a b c d))
(list-set s 2 'blah)	; (a b blah d)
s			; (a b c d)

;; destruktivni modifikace prvku
(define list-set!
  (lambda (l index value)
    (let iter ((l l)
	       (i 0))
      (if (= i index)
	  (set-car! l value)
	  (iter (cdr l) (+ i 1))))
    l))

(define s '(a b c d))
(list-set! s 2 'blah)	; (a b blah d)
s			; (a b blah d)

;;; DESTRUKTIVNI SPOJENI SEZNAMU
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; klasicky nedestruktivni (funkcionalni) append2
(define append2
  (lambda (l1 l2)
    (if (null? l1)
	l2
	(cons (car l1) (append2 (cdr l1) l2)))))

(define x '(a b c))
(define y '(10 20))
(append2 x y)	; (a b c 10 20)
x		; (a b c)
y		; (10 20)

;; destruktivne spoj dva seznamy
(define append2!
  (lambda (l1 l2)
    (if (null? l1)
	l2
	(let iter ((l l1))
	  (if (null? (cdr l))
	      (begin
		(set-cdr! l l2)
		l1)
	      (iter (cdr l)))))))

; pri spojeni seznamu dochazi k mutaci prvniho argumentu:
(define x '(a b c))
(define y '(10 20))
(append2! x y)	; (a b c 10 20)
x		; (a b c 10 20) !
y		; (10 20)

; neplati v pripade prazdneho seznamu (neni to par)
(define x '())
(define y '(10 20))
(append2! x y)	; (10 20)
x		; '()
y		; (10 20)

;; muzeme pochopitelne rozsirit na libovolne argumenty:
(define append!
  (lambda lists
    (foldr append2! '() lists)))

; dochazi k destrukci vsech krome posledniho
; vsechny seznamy se postupne ,,povazou'' (prazdne jsou ignorovany)
(define a '(a b c))
(define b '(#t #f))
(define c '(2 4 6 8))
(define d '(foo bar baz))
(append! a b c d)	; (a b c #t #f 2 4 6 8 foo bar baz)
a	; (a b c #t #f 2 4 6 8 foo bar baz)
b	; (#t #f 2 4 6 8 foo bar baz) 
c	; (2 4 6 8 foo bar baz)
d	; (foo bar baz)

;;; DESTRUKTIVNI PRIDAVANI PRVKU
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; destruktivni pridavani prvku do NEPRAZDNEHO seznamu
;; pro prazdny seznam nefunguje (nema smysl, neni mutovatelny)
(define list-insert!
  (lambda (l index value)
    (if (= index 0)	; vkladani na zacatek (specialni pripad)
	(begin
	  (set-cdr! l (cons (car l) (cdr l)))
	  (set-car! l value))
	(let iter ((l l)
		   (index index))	; vkladani doprostred
	  (if (= index 1)
	      (set-cdr! l (cons value (cdr l)))
	      (iter (cdr l) (- index 1)))))))

; ukazky pouziti
(define s '(a b c d))
(list-insert! s 0 'blaha)
(list-insert! s 1 'blahb)
(list-insert! s 2 'blahc)
(list-insert! s 3 'blahd)
(list-insert! s 4 'blahe)
s

;; Pozn: Nefunguje pro PRAZDNY SEZNAM: () neni (mutovatelny) par

;;; DESTRUKTIVNI MAZANI PRVKU
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; uladit + zkontrolovat
(define list-delete!
  (lambda (l index)
    (if (= index 0)
	(begin
	  (set-car! s (cadr s))		; spec. pripad: mazani z prvni pozice
	  (set-cdr! s (cddr s)))
	(let iter ((l l)		; mazani ze zbytku seznamu
		   (index index))
	  (if (= index 1)
	      (set-cdr! l (cddr l))
	      (iter (cdr l) (- index 1)))))))

; ukazky pouziti
(define s '(a b c d))
(list-delete! s 0)
(list-delete! s 1)
(list-delete! s 2)
(list-delete! s 3)
s

;; Pozn: Nefunguje pro JEDNOPRVKOVY SEZNAM: nelze jej ,,zkratit na prazdny''

;; Pozn: Vec ke vsimnuti: pri destruktivni praci se seznamy je potreba
;;       rozlisovat mnohem vice ,,specialnich pripadu'', to jsme pri ciste
;;       funkcionalni manipulaci se seznamy delat vubec nemuseli

;;; IMPLEMENTACE MUTOVATELNE FRONTY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; vytvor prazdnou frontu
(define make-queue
  (lambda ()
    (cons '() '())))

;; testuj, zda-li je dana fronta prazdna
(define empty-queue?
  (lambda (queue)
    (and (null? (car queue))
	 (null? (cdr queue)))))

;; vrat prvek na vrcholu fronty
(define queue-get caar)

;; vloz prvek na konec fronty
(define queue-insert!
  (lambda (queue elem)
    (if (empty-queue? queue)
	(begin
	  (set-car! queue (cons elem (car queue)))
	  (set-cdr! queue (car queue)))
	(begin
	  (set-cdr! (cdr queue) (list elem))
	  (set-cdr! queue (cddr queue))))))

;; smaz prvek z vrcholu fronty
(define queue-delete!
  (lambda (queue)
    (if (not (empty-queue? queue))
	(begin
	  (set-car! queue (cdar queue))
	  (if (null? (car queue))
	      (set-cdr! queue '()))))))

(define q (make-queue))
q
(queue-insert! q 10)
(queue-insert! q 20)
(queue-insert! q 30)
q
(queue-get q)
(queue-delete! q)
(queue-insert! q 40)
q
(queue-delete! q)
(queue-delete! q)

;;; CYKLICKE STRUKTURY
;;;;;;;;;;;;;;;;;;;;;;

;; co myslime tim cyklickym seznamem?
;; vyjdeme z prvniho paru, jdeme po smeru sipek + dostaneme se dvakrat,
;;  trikrat, ... n-krat do tehoz paru

;; nejjednodussi priklad
(define a '(ahoj))
(set-cdr! a a)
a	; DrScheme vypise: #0=(ahoj . #0#)
	; ostatni interprety (a a a ... ,,do nekonecna''

;; Poznamka: zajimavy rys, u predchoziho seznamu muzeme ,,do nekonecna''
;;           opakovat cdr (vrat zbytek bez prvniho prvku), anich bychom
;;           se kdy dostali ke ()
;;           To ma sve dusledky: tradicni operace nad seznamy selhavaji

;; kdybychom meli length napsany:
(define length
  (lambda (l)
    (if (null? l)
	0
	(+ 1 (length (cdr l))))))

;; pak bude (length a) cyklit
;; vetsina interpretu umi cyklicne seznamy odhalit:
(length a)	; length: exp. arg. of type <prop. list>; given #0=(ahoj . #0#)

;; dalsi priklad
(define a '(1 2 3))
(set-cdr! (cddr a) a)
a	; DrScheme vypise #0=(1 2 3 . #0#)
	; to jest jedna se o seznam (1 2 3 1 2 3 1 2 3 ...

;; jiny priklad
(define a '(10 20 30))
(set-car! (cdr a) a)
a		; DrScheme vypise: #0=(10 #0# 30)
(length a)	; 3

;; tj. a je triprvkovy seznam (10 (10 (10 (10 ... 30) 30) 30) 30)
;; dochazi zde k ,,nekonecnemu rekurzivnimu vnoreni seznamu ,,do sebe sama''

;; posledni priklad
(define a '(#f))
(set-car! a a)
a		; #0=(#0#)
(length a)	; 1

;; tj. a je jednoprvkovy seznam, ktery obsahuje (pouze) sebe sama

;; POJEM: REFERENCNE TRANSPARENTNI JAZYK

;;; CYKLICKE LINEARNI SEZNAMY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; zacykleni linearniho seznamu (vraci nedef. hodnotu)
;; do posledniho paru misto () vlozi ukazatel na prvni par
(define cycle!
  (lambda (l)
    (let iter ((aux l))
      (if (null? (cdr aux))
	  (set-cdr! aux l)
	  (iter (cdr aux))))))

(define s '(a b c d e))
(cycle! s)
s		; #0=(a b c d e . #0#)

;; OTAZKA: jek detekovat cyklus v seznamu?
;; soustredime se nejprve na linearni cyklicky seznam

;; naivni (nefunkcni) reseni
;; ma zasadni problem, dokaze pouze zjistit, ze cyklicky NENI
;; pro cyklicke seznamy vypocet nikdy neskonci
(define cycle?
  (lambda (l)
    (if (null? l)
	#f
	(cycle? (cdr l)))))

; problem:
(cycle? '(a b c d))	; #f .. spravna odpoved
; (cycle? zacykleny-s)  ; .. odpovedi se nikdy nedockame

;; potrebujeme porovnavat fyzicke umisteni paru v pameti
;; je jasne, ze predikal equal? nam nijak nepomuze, protoze:
(define a (cons 1 2))
(define b (cons 1 2))
(equal? a b)	; #t ... ackoliv a b jsou jinde v pameti
; test:
(set-car! a 10)
a	; (10 . 2)
b	; (1 . 2)

;; eq?:    je #t na cislech p.k. maji shodnou reprezentaci
;;               na symbolech p.k. jmenuji se stejne
;;               na parech p.k. maji stejne ulozeni v pameti
;; POZOR: predikat je castecne implementacne zavisly
;;        lze jej chapat jako zakladni predikat rovnosti
(eq? 1.2 1.2)		    ; #f
(eq? 2 2)		    ; #t
(eq? 'ahoj 'ahoj)  	    ; #t
(eq? (cons 1 2) (cons 1 2)) ; #f

;; eqv?:   je #t na cislech p.k. jsou numericky rovne
;;               na symbolech p.k. jmenuji se stejne
;;               na parech p.k. maji stejne ulozeni v pameti
;; POZOR: predikat je castecne implementacne zavisly
(eqv? 1.2 1.2)		     ; #t
(eqv? 2 2)		     ; #t
(eqv? 'ahoj 'ahoj)  	     ; #t
(eqv? (cons 1 2) (cons 1 2)) ; #f

;; eqv? bychom mohli naprogramovat
(define eqv?
  (lambda (x y)
    (if (and (number? x) (number? y))
	(= x y)
	(eq? x y))))

;; presneji: R5RS jeste vyzaduje, pokud se jedna o dve cisla,
;; aby byla obe bud v presne reprezentaci, nebo v semilogaritmicke reprezentaci
;; pomocna procedura: textuje zda-li oba elementy jsou stejneho typu

;; to jest:
(eqv? 2 2.0)	; #f
(= 2 2.0)	; #t (eqv? se tedy na cislech chova jinak nez =)

;; pomocna procedura testujici, zda-li maji oba elementy tutez vlastnost
(define both
  (lambda (type? x y)
    (and (type? x) (type? y))))

;; novy eqv?
(define eqv?
  (lambda (x y)
    (if (and (both number? x y)
	     (or (both exact? x y)
		 (both inexact? x y)))
	(= x y)
	(eq? x y))))

;; equal?: na cislech a symbolech se chova stejne jako eqv?
;;         pary jsou rekurzivne prochazeny a testovana jejich ekvivalence: 
(equal? 1.2 1.2)	       ; #t
(equal? 2 2)		       ; #t
(equal? 'ahoj 'ahoj)  	       ; #t
(equal? (cons 1 2) (cons 1 2)) ; #t

;; implementace equal?
(define equal?
  (lambda (x y)
    (cond ((both number? x y) (eqv? x y))
	  ((both symbol? x y) (eqv? x y))
	  ((both pair? x y) (and (equal? (car x) (car y))
				 (equal? (cdr x) (cdr y))))
	  (else (eq? x y)))))

;; test zacyklenosti linearniho seznamu
(define cyclic?
  (lambda (l)
    (let test ((rest (if (null? l)
			 '()
			 (cdr l))))
      (cond ((null? rest) #f)
	    ((eq? rest l) #t)
	    (else (test (cdr rest)))))))

(cyclic? '())		; #f
(cyclic? '(a b c))	; #f
(define s '(a b c))
(cycle! s)
(cyclic? s)		; #t

;; odcykleni linearniho seznamu
;; spociva v rozetnuti cyklu vlozenim () misto ukazatele na pocatek
(define uncycle!
  (lambda (l)
    (let iter ((aux l))
      (if (eq? (cdr aux) l)
	  (set-cdr! aux '())
	  (iter (cdr aux))))))

; zacykleni + odcykleni seznamu
(define s '(a b c))
(cycle! s)
s	; #0=(a b c . #0#)
(uncycle! s)
s	; (a b c)

; zacyklednim + odcyklenim nemusime ziskat vychozi seznam
(define s '(a b c))
(cycle! s)
(set! s (cdr s))
s	; #0=(b c a . #0#)
(uncycle! s)
s	; (b c a)

;;; SLOZITEJSI CYKLICKE STRUKTURY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Analogicky jako existuji eq?, eqv?, equal?
;; maji sve varianty i member a assoc

;; member -- pouziva k porovnani prvku equal?
;; memv -- pouziva k porovnani prvku eqv?
;; memq -- pouziva k porovnani prvku eq?
(member '(a) '(1 2 (a) 3 4))	; ((a) 3 4)
(memq '(a) '(1 2 (a) 3 4))	; #f

;; assoc -- pouziva k porovnani klicu equal?
;; assv -- pouziva k porovnani klicu eqv?
;; assq -- pouziva k porovnani klicu eq?
(assoc '(2) '(((1) . a) ((2) . b) ((3) . c)))	; ((2) . b)
(assq '(2) '(((1) . a) ((2) . b) ((3) . c)))	; #f

;; test cyklu do hloubky -- behem sestupu seznamem si udrzujeme seznam
;; jiz navstivenych paru, procedura vyuziva memq:
(define depth-cyclic?
  (lambda (l)
    (let test ((l l)
	       (found '()))
      (if (pair? l)
	  (if (memq l found)
	      #t
	      (let ((found (cons l found)))
		(or (test (car l) found)
		    (test (cdr l) found))))
	  #f))))

; priklad pouziti
(define s '(a b c d e f))
(set-car! (cdddr s) (cdr s))
s	; (a . #0=(b c #0# e f))
(length s)	; 6 takze neni cyklicka v tom predchozim (jednoduchem) pojeti
; avsak:
(depth-cyclic? s)	; #t

;; Poznamka: vratme se opet k nasi implementaci equal?
(define equal?
  (lambda (x y)
    (cond ((both number? x y) (eqv? x y))
	  ((both symbol? x y) (eqv? x y))
	  ((both pair? x y) (and (equal? (car x) (car y))
				 (equal? (cdr x) (cdr y))))
	  (else (eq? x y)))))

;; je jasne, ze pro:
(define s '(a))
(set-cdr! s s)
; (equal? s (cons 'a s))	; cyklii
; intuitivne bychom ocekevali, ze predchozi by mel vratit #t

;; jemne upraveny equal?
;; vyuzivame toho, ze equal? je zasilenim eqv?
(define equal?
  (lambda (x y)
    (or (eqv? x y)
	(and (both pair? x y)
	     (equal? (car x) (car y))
	     (equal? (cdr x) (cdr y))))))

;; ted uz funguje:
(define s '(a))
(set-cdr! s s)
(equal? s s)		; #t
(equal? s (cons 'a s))	; #t

;; porad to ale neni ono:
(define s '(a))
(define r '(a))
(set-cdr! s s)
(set-cdr! r r)
; (equal? s r)	; bude cyklit, intuitivne bychom ocekavali #t
;; pozn. vetsina interpretu Scheme na tomhle cykli / nebo spadne
;; Common LISP taky. Cha Cha, Krupko!

; otazka, je tento prblem (vubec) resitelny?
; resitelny je, ale je to drsne (NEPOUZITELNA VYPOCETNI SLOZITOST):
; prevedeni na problem ekvivalence dvou regularnich jazyku
; pary = stavy KD automatu
; koncove stav = jeden, oznacuje atomicky prvek
; hrany: ohodnocene l,r, pripadne la, ra, kde a je nazev atomu pri kterem
; jdeme do koncoveho stavu
; dva cyklicke seznamy jsou ekvivalentni if{}f
; L(R_1) = L(R_2) resitelny problem (i kdyz jazyky L(R_1) a L(R_2) jsou
; v pripade zacyklenych seznamu NEKONECNE)
; (JE RESITELNY, casove pozadavky O(2^n))
; rozebrat do knizky

;; OBOUSMERNE SEZNAMY: SPECIALNI CYKLICKE STRUKTURY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; (elem . (ptr1 . ptr2))
;; ptr1 .. ukazatel na predchudce
;; ptr2 .. ukazatel na naslednika

;; konstruktor obousmerneho seznamu
(define cons-dlist
  (lambda (elem dlist)
    (let ((new-cell (cons elem
			  (cons '()
				dlist))))
      (if (not (null? dlist))
	  (set-car! (cdr dlist) new-cell))
      new-cell)))

;; selektor car
(define dlist-car
  (lambda (dlist)
    (car dlist)))

;; selektor cdr (zbytek seznam bez aktualniho prvku)
(define dlist-cdr
  (lambda (dlist)
    (cddr dlist)))

;; selektor cir (predek seznamu bez aktualniho prvku)
(define dlist-cir
  (lambda (dlist)
    (cadr dlist)))

(cons 'a (cons 'b (cons 'c (cons 'd '()))))
(cons-dlist 'a (cons-dlist 'b (cons-dlist 'c (cons-dlist 'd '()))))
(define s (cons-dlist 'a (cons-dlist 'b (cons-dlist 'c (cons-dlist 'd '())))))
(dlist-car s)
(dlist-cir s)
(dlist-cdr s)
(dlist-car (dlist-cdr s))
(dlist-cir (dlist-cdr s))
(dlist-cdr (dlist-cdr s))
(dlist-car (dlist-cdr (dlist-cdr s)))
(dlist-cir (dlist-cdr (dlist-cdr s)))
(dlist-cdr (dlist-cdr (dlist-cdr s)))
(define z (dlist-cdr (dlist-cdr (dlist-cdr s))))
(dlist-car z)
(dlist-cdr z)
(dlist-car (dlist-cir z))
(dlist-car (dlist-cir (dlist-cir z)))

;; prevod obycejneho linearniho seznamu na dlist
(define list->dlist
  (lambda (l)
    (if (null? l)
	'()
	(cons-dlist (car l)
		    (list->dlist (cdr l))))))

;; hloubkovy prevod obycejneho seznamu na dlist
(define depth-list->dlist
  (lambda (l)
    (if (null? l)
	'()
	(cons-dlist (if (list? (car l))
			(depth-list->dlist (car l))
			(car l))
		    (depth-list->dlist (cdr l))))))

(depth-list->dlist '(a (b c d) (e f)))

;; prevod linearniho dlistu na klasicky seznam (vpred / vzad)
(define dlist->list
  (lambda (dl . backwards)
    (if (null? dl)
	'()
	(cons (dlist-car dl)
	      (apply dlist->list
		     (if (null? backwards)
			 (dlist-cdr dl)
			 (dlist-cir dl))
		     backwards)))))

(dlist->list (list->dlist '(a b c d)))
(dlist->list (dlist-cdr (dlist-cdr (list->dlist '(a b c d)))) #t)

;; DC: napiste mutatory: dlist-set-car!, dlist-set-cdr!, dlist-set-cir!

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; PREDAVANI ARGUMENTU PROCEDURAM
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; tohle pochopitelne nemeni hodnotu predavaneho argumentu
(define add2
  (lambda (x)
    (set! x (+ x 2))
    x))

(define val 10)
(add2 val)	; ]=> 12
val		; ]=> 10

;;; BOX (MUTOVATELNY KONTAINER NA HODNOTU)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; vytvori box -- novy objekt, ktery reaguje na dva signaly
;; set .. zapis hodnotu do boxu
;; get .. vrat hodnotu z boxu
(define make-box
  (lambda (value)
    (lambda (signal . new-value)
      (cond ((equal? signal 'get) value)
	    ((equal? signal 'set) (set! value (car new-value)))
	    (else "neznamy signal")))))

(define val (make-box 10))
(val 'get)	; ]=> 10

;; vypocti faktorial a zapis vyslednou hodnotu do argumentu
;; vraci vzdy symbol 'hotovo
(define proc
  (lambda (box n)
    (letrec ((f (lambda (n)
		  (if (= n 1)
		      1
		      (* n (f (- n 1)))))))
      (box 'set (f n))
      'hotovo)))

(proc val 20)
(val 'get)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; DALSI MUTOVATELNE ELEMENTY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; VEKTORY (analogie pole)
;;;;;;;;;;;;;;;;;;;;;;;;;;

;; vytvareni vektoru pomoci hodnot
(vector)		; #0()
(vector 10 20 30)	; #2(10 20 30)
(vector 10 10 10)	; #3(10)
(vector 'ahoj 'svete)	; #2(ahoj svete)
(vector 1 #f 'blah)	; #3(1 #f blah) ... tj. lze ukladat veci ruznych typu

;; vrat delku vektoru
(vector-length (vector))		; 0
(vector-length (vector 'a 'b 'c))	; 3

;; vytvareni vektoru o dane delce (hodnoty nespecifikovane)
;; ruzne interprety maji ruzne hodnoty
(make-vector 10)	; #10(0)

;; naplneni vektoru jednou hodnotou
(define v (make-vector 10))
(vector-fill! v 'blah)
v			; #10(blah)

;; vytvareni vektoru o dane delce s pocatecnim naplnenim
(make-vector 10 'blah)	; #10(blah)

;; ziskani hodnoty podle indexu / mutace hodnoty
(define v (vector 'a 'b 'c 'd 'e 'f))
(vector-ref v 2)	; c
(vector-set! v 2 'blah)
(vector-ref v 2)	; blah
v			; #6(a b blah d e f)

;; CVICENI:
;;;;;;;;;;;

;; list->vector
(define list->vector
  (lambda (l)
    (let ((new-vector (make-vector (length l))))
      (let iter ((l l)
		 (index 0))
	(if (null? l)
	    new-vector
	    (begin
	      (vector-set! new-vector index (car l))
	      (iter (cdr l) (+ index 1))))))))

;; vector->list
(define vector->list
  (lambda (v)
    (build-list (vector-length v)
		(lambda (i)
		  (vector-ref v i)))))

;; build-vector (analogicky jako build-list, akorat dela vektor)
(define build-vector
  (lambda (len f)
    (let ((new-vector (make-vector len)))
      (let iter ((i 0))
	(if (>= i len)
	    new-vector
	    (begin
	      (vector-set! new-vector i (f i))
	      (iter (+ i 1))))))))

;; vector-map2: jako map2, akorat pracuje s vektorem
;; verze pro jeden vektor
(define vector-map2
  (lambda (f v)
    (build-vector (vector-length v)
		  (lambda (i)
		    (f (vector-ref v i))))))

;; verze pro libovolne mnozstvi vektoru
(define vector-map
  (lambda (f . vectors)
    (build-vector (vector-length (car vectors))
		  (lambda (i)
		    (apply f (map (lambda (v)
				    (vector-ref v i))
				  vectors))))))

;; kopie vektoru, od - do
(define vector-copy
  (lambda (v start end)
    (build-vector (+ 1 (- end start))
		  (lambda (i)
		    (vector-ref v (+ i start))))))

;; muzeme udelat verze cons, car a cdr pro vektory

;; analogie cons
(define cons-vector
  (lambda (elem v)
    (build-vector (+ (vector-length v) 1)
		  (lambda (i)
		    (if (= i 0)
			elem
			(vector-ref v (- i 1)))))))

;; analogie car
(define vector-first
  (lambda (v)
    (vector-ref v 0)))

;; analogie cdr
(define vector-rest
  (lambda (v)
    (build-vector (- (vector-length v) 1)
		  (lambda (i)
		    (vector-ref v (+ i 1))))))

;;;; POROVNANI SLOZITOSTI (ZHRUBA)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; build-list  O(n)  |  build-vector   O(n)
; car         O(1)  |  vector-car     O(1)
; cdr         O(1)  |  vector-cdr     O(n)
; cons        O(1)  |  cons-vector    O(n)
; length      O(n)  |  vector-length  O(1)
; list-ref    O(n)  |  vector-ref     O(1)
; list-set!   O(n)  |  vector-set!    O(1)
; map         O(n)  |  vector-map     O(n)

;; STRUKTURY (de facto totez jako vektory)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; (define-struct ... ma DrScheme, neni v R5RS, nebudeme se tim zabyvat
; mozna ani neukazovat, jen ,,teoreticky slajd''
