
;;  Paradigmata Programovani 2A, PRIKLADY
;;  Copyright (C) 2008  Vilem Vychodil, <vilem.vychodil@upol.cz>
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; version 2 of the License.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;; priklady maker v C: #define MAKRO(__neco1,__neco2) ...
;; priklady maker v M4: maso ... jazyk zalozeny na makroexpanzi

;; poradi makro -- procedura, pokud procedura pouziva makro a my jej zmenime,
;; musime znova ,,prelozit proceduru'' (po makroexpanzi)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; KVAZIKVOTOVANI (PRIPOMINAME)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; ,,vylepsena verze'' klasickeho kvotovani
`(1 2 3 4 5)		; (1 2 3 4 5)
`(1 (+ 2 3) 4 5)	; (1 (+ 2 3) 4 5)
`(1 ,(+ 2 3) 4 5)	; (1 5 4 5)
`(1 ((,(+ 2 3))) 4 5)	; (1 ((5)) 4 5)

(define s '(a b c))
`(1 ,s 2)		; (1 (a b c) 2)
`(1 ,@s 2)		; (1 a b c 2)

;; jak se chova vzhledem ke quote:
`(1 '2 3)		; (1 (quote 2) 3)
`'(1 2 3)		; (quote (1 2 3))
'`(1 2 3)		; (quasiquote (1 2 3))

;; vnorene kvazikvotovani
`(1 `(2 3))		; (1 (quasiquote (2 3)))
`(1 `(,(+ 1 2) 3))	; (1 (quasiquote ((unquote (+ 1 2)) 3)))
`(1 ,`(,(+ 1 2) 3))	; (1 (3 3))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; MAKRA (ZAKLADNI VECI)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; PRIKLAD S MODIFIKOVANYM `IF'
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; chceme upravit `if' tak, aby pri absenci alternativniho vyrazu vracel #f

;; nyni mame:
(if (= 1 2) 'blah)	; nedef. hodnota

;; chceme:
(new-if (= 1 2) 'blah)	; #f

;; nabizi se ,,vyresit pomoci procedury'':
(define new-if
  (lambda (elem1 elem2)
    (if elem1 elem2 #f)))

;; pri volani new-if je vzdy vyhodnocen i druhy argument, to jest
(new-if #f blah-blah)	; konci chybou

;; POTREBUJEME: behem vyhodnocovani kazdy vyraz
(new-if expr1 expr2)
;; nahradit vyrazem
(if expr1 expr2 #f)
;; bez toho aniz by se vyhodnocovaly expr1 a expr2

;; transformacni procedura
(define new-if-transformer
  (lambda (test expr . alt)
    (list 'if test expr
	  (if (null? alt)
	      #f
	      (car alt)))))

; priklad:
(new-if-transformer 'expr1 'expr2 'expr3)	; (if expr1 expr2 expr3)
(new-if-transformer 'expr1 'expr2)		; (if expr1 expr2 #f)
(new-if-transformer '(even? x) '(+ x 1) #f)	; (if (even? x) (+ x 1) #f)

; rucni reseni
(eval (new-if-transformer '(even? x) '(+ x 1) #f))

; chova se spatne k lexikalnimu rozsahu
(let ((y 10))
  (eval (new-if-transformer '(even? y) '(+ y 1))))

(let ((y 10))
  (eval (new-if-transformer '(even? y) '(+ y 1))
	(the-environment)))

;; vytvoreni MAKRA pomoci transformacni procedury
(define-macro new-if
  (lambda (test expr . alt)
    (list 'if test expr
	  (if (null? alt)
	      #f
	      (car alt)))))

;; vytvoreni MAKRA pomoci transformacni procedury
(define-macro new-if
  (lambda (<test> <expr> . <alt>)
    (list 'if <test> <expr>
	  (if (null? <alt>)
	      #f
	      (car <alt>)))))

(new-if 1 2 3)			; 2
(new-if (even? 2) 10 20)	; 10
(new-if #f blah-blah)		; #f

;; kratsi reseni pomoci kvazikvotovani
(define new-if-transformer
  (lambda (test expr . alt)
    `(if ,test
	 ,expr
	 (begin #f ,@alt))))

(new-if-transformer 'expr1 'expr2 'expr3)  ; (if expr1 expr2 (begin #f expr3))
(new-if-transformer 'expr1 'expr2)	   ; (if expr1 expr2 (begin #f))
(new-if-transformer '(even? x) '(+ x 1) #f); (if (even? x) (+ x 1) (begin #f #f))

;; nove makro
(define-macro new-if
  (lambda (test expr . alt)
    `(if ,test
	 ,expr
	 (begin #f ,@alt))))

;; nove makro
(define-macro new-if
  (lambda (<test> <expr> . <alt>)
    `'(if ,<test>
	 ,<expr>
	 (begin #f ,@<alt>))))

(let ((x 10))
  (new-if (even? x) (+ x 1)))

(if (even? x) (+ x 1) (begin #f))

11

;; JAK LADIT MAKRO?

;; chceme potlacit fazi ,,vyhodnoceni transformovaneho kodu''
;; muzeme zaridit pomoci kvotovani

;; 1) slozitejsi
(define-macro new-if
  (lambda (test expr . alt)
    (list 'if test expr
	  (if (null? alt)
	      #f
	      (car alt)))))

(new-if 1 2 3)			; (if 1 2 3)
(new-if (even? 2) 10 20)	; (if (even? 2) 10 20)
(new-if #f blah-blah)		; (if #f blah-blah #f)

;; 2) jednodussi
(define-macro new-if
  (lambda (test expr . alt)
    `'(if ,test ,expr (begin #f ,@alt))))

(new-if 1 2 3)			; (if 1 2 (begin #f 3))
(new-if (even? 2) 10 20)	; (if (even? 2) 10 (begin #f 20))
(new-if #f blah-blah)		; (if #f blah-blah (begin #f))

;;; PRIKLAD S `AND2'
;;;;;;;;;;;;;;;;;;;;

;; predpokladejme, ze mame pouze if, nemame `cond' ani `and' ani `or'
;; chceme vytvorit and2 (and dvou argumentu) vracejici #t nebo #f

; reseni pomoci procedury:
(define and2
  (lambda (elem1 elem2)
    (if elem1
	(if elem2
	    #t
	    #f)
	#f)))

; priklad:
(and2 (even? y) (<= x 3))
	    
;; pokud je navic PODMINKA:
;; pokud je prvni vyraz pravdivy, druhy se nevyhodnocuje
;; pak jiz nemuzeme vyresit pomoci procedury!
(and2 #f blah-blah)	; #f

;; POTREBUJEME: behem vyhodnocovani kazdy vyraz
(and2 expr1 expr2)
;; nahradit vyrazem
(if expr1
    (if expr2
	#t
	#f)
    #f)
;; bez toho aniz by se vyhodnocovaly expr1 a expr2

;; reseni pomoci makra:
(define-macro and2
  (lambda (expr1 expr2)
    `(if ,expr1
	 (if ,expr2
	     #t
	     #f)
	 #f)))

; ukazky pouziti:
(and2 #f 2)

(and2 1 (+ 1 2))
(if 1 (if (+ 1 2) #t #f) #f)
#t

(and2 #f blah-blah)
(if #f (if blah-blah #t #f) #f)
#f

(and2 #t #f)
(if #t (if #f #t #f) #f)
#f

;;; PRIKLAD: DALSI MODIFIKACE IF
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; if*, ktery pracuje stejne jako if, ale umoznuje v duhem a tretim vyrazu
;; pouzivat symbol $result, ktery bude vzdy navazany na vysledek vyhodnoceni
;; prvniho vyrazu

;; prakticka vec, misto:
(if (member 'b '(a b c d))
    (list 'nalezen (member 'b '(a b c d)))
    'blah)

;; staci napsat:
(if* (member 'b '(a b c d))
     (list 'nalezen $result)
     'blah)			; (nalezen (b c d))

;; RESENI: behem vyhodnocovani kazdy vyraz
(if* expr1 expr2 expr3)
;; porebujeme nahradit vyrazem:
(let (($result expr1))
  (if $result expr2 expr3))
;; opet bez vyhodnocovani expr1 az expr3

;; transformator
(define if*-transformer
  (lambda (test expr . alt)
    `(let (($result ,test))
       (if $result
	   ,expr
	   ,@alt))))

(if*-transformer '(member 'b '(a b c d))
		 '(list 'nalezen $result)
		 ''blah)

; Vysledek:
(let (($result (member (quote b) (quote (a b c d)))))
  (if $result
      (list (quote nalezen) $result)
      (quote blah)))

;; if, ktery umoznuje pristupovat v vysledku podminky
(define-macro if*
  (lambda (test expr . alt)
    `(let (($result ,test))
       (if $result
	   ,expr
	   ,@alt))))

;; ukazky pouziti
(if* 1 2 3)
(if* 1 $result 3)
(if* #f blah 3)
(if* #f 1)
(if* (member 'b '(a b c d))
     (list 'nalezen $result)
     'blah)

(if* 1 2 3)
(let (($result 1)) (if $result 2 3))
2

(if* 1 $result 3)
(let (($result 1)) (if $result $result 3))
1

(if* (member 'b '(a b c d))
     (list 'nalezen $result)
     'blah)
(let (($result (member (quote b)
		       (quote (a b c d)))))
  (if $result
      (list (quote nalezen) $result)
      (quote blah)))
(nalezen (b c d))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; IF pomoci COND a COND pomoci IF
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; if pomoci cond
(define-macro if
  (lambda (test expr alt)
    `(cond (,test ,expr)
	   (else ,alt))))

;; if pomoci cond (bez nutnosti mit alternativni vetev)
(define-macro if
  (lambda (test expr . alt)
    `(cond (,test ,expr)
	   (else (begin #f ,@alt)))))

;; podobne jako predchozi, ale vracime nedef. hodnotu
;; zde pripomenout co dela ``prazdny cond''
(define-macro if
  (lambda (test expr . alt)
    `(cond (,test ,expr)
	   (else (begin (cond) ,@alt)))))

;; zakladni cond pomoci if (pomoci rekurzivniho vnoreni)
(define-macro cond
  (lambda clist
    (let dive-ifs ((clist clist))
      (if (null? clist)
	  `(if #f #f)
	  (if (equal? (caar clist) 'else)
	      (cadar clist)
	      `(if ,(caar clist)
		   ,(cadar clist)
		   ,(dive-ifs (cdr clist))))))))

(cond ((= x 3) 'blah)
      ((> x 10) (+ 1 x))
      ((prop? x y) (list x y))
      (else (f 20)))

(if (= x 3)
    (quote blah)
    (if (> x 10)
	(+ 1 x)
	(if (prop? x y)
	    (list x y)
	    (f 20))))

(if (= x 3)
    (quote blah)
    (cond ((> x 10) (+ 1 x))
	  ((prop? x y) (list x y))
	  (else (f 20))))

;; zakladni cond pomoci if (rekurzivni makro)
(define-macro cond
  (lambda clist
    (if (null? clist)
	`(if #f #f)
	(if (equal? (caar clist) 'else)
	    (cadar clist)
	    `(if ,(caar clist)
		 ,(cadar clist)
		 (cond ,@(cdr clist)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ROZSIRENA VERZE DEFINE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; define jsme zatim pozuivali pouze ve tvaru (define symbol vyraz)
(define symbol 10)

;; v R5RS je zaveden define taky ve tvaru
;; (define (symbol argumenty...) vyrazy...)

;; faktorial
(define (f n)
  (if (= n 1)
      1
      (* n (f (- n 1)))))

;; nepovinne argumenty
(define (f x y . args)
  (list x y args))

(f 1 2 3)	; (1 2 (3))

;; libovolne argumenty
(define (f . args)
  args)

(f 1 2 3)	; (1 2 3)

;; pokud by nas interpret nedisponoval rozsirenym define, muzeme jej vyrobit:
(define-macro def
  (lambda (first . args)
    (if (symbol? first)
	`(define ,first ,@args)
	`(define ,(car first)
	   (lambda ,(cdr first)
	     ,@args)))))

(def x 10)
x	; 10

(def (f n)
  (if (= n 1)
      1
      (* n (f (- n 1)))))
(f 6)	; 720

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; SPECIALNI FORMY AND a OR
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; specialni forma ``and'' pomoci if (zakladni verze)
(define-macro and
  (lambda args
    (if (null? args)
	#t
	`(if ,(car args) 
	     (and ,@(cdr args))
	     #f)))))

(and 1 2 3)	; (if 1 (and 2 3) #f)
(and 2 3)       ; (if 2 (and 3) #f)
(and 2)         ; (if 2 (and) #f)
(and)           ; #t

;; specialni forma ``and'' pomoci if (zobecnene pravdivostni hodnoty)
(define-macro and
  (lambda args
    (if (null? args)
	#t
	(if (null? (cdr args))
	    (car args)
	    `(if ,(car args) 
		 (and ,@(cdr args))
		 #f)))))

(and)		; #t
(and 1 2 3)	; 3
(and 1 #f 3)	; #f

;; specialni forma ``or'' (zakladni verze)
(define-macro or
  (lambda args
    (if (null? args)
	#f
	`(if ,(car args) 
	     #t
	     (or ,@(cdr args))))))

(or)		; #f
(or 1 2 3)	; #t (melo by byt 1)
(or #f 2 3)	; #t (melo by byt 2)

;; specialni forma ``or'' pomoci if (rozsirena verze s hackem)
(define-macro or
  (lambda args
    (if (null? args)
	#f
	(if (null? (cdr args))
	    (car args)
	    `(if ,(car args)
		 ,(car args)
		 (or ,@(cdr args)))))))

(or)		; #f
(or 1 2 3)	; 1
(or #f 2 3)	; 2

;; POZOR: ma to hacek

;; klasicky ``or''
(let ((x 0))
  (or (begin (set! x (+ x 1))
	     x)
      blah))		; ]=> 1

;; nas ``or''
(let ((x 0))
  (or (begin (set! x (+ x 1))
	     x)
      blah))		; ]=> 2

;; upravime or tak, aby k vyhodnoceni vyrazu dochazelo jen jednou
(define-macro or
  (lambda args
    (if (null? args)
	#f
	(if (null? (cdr args))
	    (car args)
	    `(let ((result ,(car args)))
	       (if result
		   result
		   (or ,@(cdr args))))))))

;; nyni uz jsme predchozi problem odstranili, ...

(let ((x 0))
  (or (begin (set! x (+ x 1))
	     x)
      blah))		; ]=> 1

;; ..., ale zadelali jsme na dalsi:

;; klasicky ``or''
(let ((result 10))
  (or #f result))	; ]=> 10

;; nas ``or''
(let ((result 10))
  (or #f result))	; ]=> #f

(let ((result #f)) (if result result (or result)))

;; predchozimu efektu se rika ,,symbol (variable) capture''
;; na pristi prednasce si ukazeme ciste reseni

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; LET, LET* a LETREC
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; zakladni let umoznujici vazat hodnoty
(define-macro let
  (lambda (assgn . body)
    `((lambda ,(map car assgn)
        (begin ,@body))
      ,@(map cadr assgn))))

(let ((x 10)
      (y (+ x 1)))
  (list x y))

((lambda (x y) (begin (list x y))) 10 (+ x 1))

;; vcetne pojmenovaneho letu
(define-macro let
  (lambda args
    (if (symbol? (car args))
	;; pojmenovany let
	`((lambda ()
	    (define ,(car args)
	      (lambda ,(map car (cadr args))
		(begin ,@(cddr args))))
	    (,(car args) ,@(map cadr (cadr args)))))
	;; nepojmenovany let
	`((lambda ,(map car (car args))
	    (begin ,@(cdr args)))
	  ,@(map cadr (car args))))))

(let f ((x 10)) (+ x 1))

;; let* (pomoci rekurzivniho vnoreni)
(define-macro let*
  (lambda (assgn . body)
    (define iter
      (lambda (assgn)
	(if (null? assgn)
	    `((lambda () ,@body))
	    `((lambda (,(caar assgn))
		,(iter (cdr assgn)))
	      ,(cadar assgn)))))
    (iter assgn)))

(let* ((x 10)
       (y (+ x 1)))
  (list x y))

((lambda (x)
   ((lambda (y)
      ((lambda () (list x y))))
    (+ x 1)))
 10)

;; let* (jako rekurzivni makro)
(define-macro let*
  (lambda (assgn . body)
    (if (null? assgn)
	`((lambda () ,@body))
        `((lambda (,(caar assgn))
	    (let* ,(cdr assgn)
	      ,@body))
	  ,(cadar assgn)))))

;; let* (jako rekurzivni makro pomoci jiz vytvoreneho let)
(define-macro let*
  (lambda (assgn . body)
    (if (null? assgn)
	`((lambda () ,@body))
        `(let (,(car assgn))
	   (let* ,(cdr assgn)
	     ,@body)))))

;; POZOR, kdybychom misto (lambda () ,@body) pouzili begin:
(define-macro let*
  (lambda (assgn . body)
    (if (null? assgn)
	`(begin ,@body)
        `(let (,(car assgn))
	   (let* ,(cdr assgn)
	     ,@body)))))

;; tak bude mit jiny vyznam, viz: (let* () (define x 10))

;; letrec: pomoci set!
(define-macro letrec
  (lambda (assgn . body)
    `((lambda ,(map car assgn)
        ,@(map (lambda (i)
                 `(set! ,(car i) ,(cadr i)))
               assgn)
        ,@body)
      ,@(map (lambda (i) #f) assgn))))

;; letrec: pomoci define (jednodussi)
(define-macro letrec
  (lambda (assgn . body)
    `((lambda ()
        ,@(map (lambda (i)
                 `(define ,(car i) ,(cadr i)))
               assgn)
        ,@body))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; DALSI (NASE) DVE VYLEPSENI FORMY LET: LET-VALUES a LETREF
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; udelejte formu `let-values', ktere umi rozpakovat vic hodnot
(define-macro let-values
  (lambda (assgn . body)
    `(apply
      (lambda
	  ,(apply append
		  (map (lambda (x)
			 (if (list? (car x))
			     (car x)
			     (list (car x))))
		       assgn))
        (begin ,@body))
      (append ,@(map (lambda (x)
		       (if (list? (car x))
			   (cadr x)
			   `(list ,(cadr x))))
		     assgn)))))

; priklad pouziti let-values:
(let ((seznam `("Vilem" 100 blah ,(+ 1 2))))
  (let-values ((blah 10)
	       ((name value next comment) seznam)
	       ((v n c) (cdr seznam))
	       (x (+ 10 20)))
    (list blah name value next comment v n c x)))

;; udelejte formu `letref', ktera umi znovu refreshovat puvodni hodnoty symbolu
;; procedura ma opet neprijemny efekt: pri refreshi vyhodnocuje dvakrat
(define-macro letref
  (lambda (bindings . body)
    `((lambda ()
	,@(map (lambda (b)
		 `(define ,(car b) ,(cadr b)))
	       bindings)
	(define refresh
	  (lambda (symbol)
	    (cond ,@(map (lambda (x)
			   `((equal? symbol ',(car x)) 
			     (set! ,(car x) ,(cadr x))))
			 bindings))))
	(begin ,@body)))))

;; testovaci kod
(letref ((x 10)
	 (f (lambda (n)
	      (if (= n 0)
		  1
		  (* n (f (- n 1))))))
	 (y 100))
  (display x)	; 10
  (newline)
  (set! x (f 20))
  (display x)	; 2432902008176640000
  (newline)
  (refresh 'x)
  (display x)	; 10
  (newline)
  #f)	; #f

;; ukazka, kde to hapruje:
(letref ((i 0)
	 (x (begin (display "VOLANA")
		   (set! i (+ i 1))
		   i)))
  (display (list i x))
  (display (list i x))
  (set! x 'blah)
  (display (list i x))
  (refresh 'x)
  (display (list i x))
  (display (list i x))
  #f)

;; predchozi vadu na krase muzeme predelat preprogramovanim `refresh'
(define-macro letref
  (lambda (bindings . body)
    `((lambda ()
	,@(map (lambda (b)
		 `(define ,(car b) ,(cadr b)))
	       bindings)
	(define refresh
	  (let ((mem (list
	    ,@(map (lambda (x)
		     `(cons ',(car x) ,(car x)))
		   bindings))))
	    (lambda (symbol)
	      (cond ,@(map (lambda (x)
	        `((equal? symbol ',(car x)) 
		  (set! ,(car x)
			(cdr (assoc ',(car x) mem)))))
			   bindings)))))
	(begin ,@body)))))

;; ted uz je OK:
(letref ((i 0)
	 (x (begin (display "VOLANA")
		   (set! i (+ i 1))
		   i)))
  (display (list i x)) ; (1 1)
  (set! x 'blah)
  (display (list i x)) ; (1 blah)
  (refresh 'x)         ; refreshne x
  (display (list i x)) ; (2 2)
  #f)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; MAKRO PRO VYTVARENI REKURZIVNICH PROCEDUR BEZ POUZITI DEFINE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; makro pro vytvareni rekurzivnich procedur bez define
;; zakukli y-kombinator
(define-macro procedure
  (lambda (args . body)
    `(lambda ,args
       ((lambda (y)
	  (y y ,@args))
	(lambda (self* ,@args)
	  ,@body)))))

(define-macro self
  (lambda args
    `(self* self* ,@args)))

(let ((f (procedure (n)
	   (if (= n 1)
	       1
	       (* n (self (- n 1)))))))
  (map f '(1 2 3 4 5 6 7)))

;; ma malou vadu na krase: self se ,,uvnitr nechova jako procedura'', viz:

(let ((f (procedure (x)
	   (if (list? x)
	       (apply + (map self x))
	       1))))
  (f '(a ((b c) ((d))) ((e) f))))	; error (self neni procedura)

;; problem muzeme odstranit necekane elegantne:
(define-macro procedure
  (lambda (args . body)
    `(lambda ,args
       ((lambda (y)
	  (y y ,@args))
	(lambda (self ,@args)
	  (let ((self (lambda ,args (self self ,@args))))
	    ,@body))))))

(let ((f (procedure (n)
	   (if (= n 1)
	       1
	       (* n (self (- n 1)))))))
  (map f '(1 2 3 4 5 6 7)))

(let ((f (procedure (x)
	   (if (list? x)
	       (apply + (map self x))
	       1))))
  (f '(a ((b c) ((d))) ((e) f))))
