
;;  Paradigmata Programovani 2A, PRIKLADY
;;  Copyright (C) 2008  Vilem Vychodil, <vilem.vychodil@upol.cz>
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; version 2 of the License.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; generovane (bezejmenne) symboly + vyresit OR
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; pojmenovane symboly
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; pojmenovany symbol = element jazyka
;; symbol, ktery je identifikovatelny svym ,,jmenem''

;; Pr.:
'ahoj	; ]=> element ``symbol'', ktery ma jmeno "ahoj"
'blah   ; ]=> element ``symbol'', ktery ma jmeno "blah"

(equal? 'ahoj 'blah)	; #f
(equal? 'ahoj 'ahoj)	; #t

;; to jest jmenem "ahoj" identifikujeme tentyz symbol (chapany jako elem. j.)
;; to je prakticke: v prostredich jsou na symboly navazane dalsi elementy,
;; je proto prakticke, aby se symboly chovaly jako "jmena"

;; ovsem, nekdy se hodi mit k dispozici (novy) symbol, ja ktery se nikdo
;; jeho jmenem odkazat nemuze

;; procedura: gensym (bez argumentu) -- vraci nove vygenerovany symbol,
;;            ktery nema jmeno

(gensym)			; neco ty vypise, ale je jedno co to je
(symbol? (gensym))		; #t
(equal? (gensym) (gensym))	; #f

(define s (gensym))
(equal? s s)		; #t ... kazdy bezejmenny symbol je roven (eq?)
			; sam sobe a zadnamu jimenu symbolu

;; i kdyby interpret dva generovane symboly ,,vypisoval stejne''
;; nejde o tyz symbol!

;; v nasledujicim ukazkovem makru dochazi k zachyceni symbolu `curval':
(define-macro capture
  (lambda body
    `(let ((curval 100))
       ,@body)))

(let ((curval 10))
  (capture
   (display "Hodnota: ")
   (display curval)
   (newline)
   (+ curval 1)))	; ]=> 101

; prepis vypada takto
(let ((curval 100))
  (display "Hodnota: ")
  (display curval)
  (newline)
  (+ curval 1))

;; v nasledujicim (upravenem) makru jiz nikoliv:
(define-macro no-capture
  (lambda body
    (let ((new-unnamed-symbol (gensym)))
      `(let ((,new-unnamed-symbol 100))
	 ,@body))))

(let ((curval 10))
  (no-capture
   (display "Hodnota: ")
   (display curval)
   (newline)
   (+ curval 1)))	; ]=> 100

; prvni faze:
(let ((<generated-symbol> 100))
  (display "Hodnota: ")
  (display curval)
  (newline)
  (+ curval 1))

;; POZOR: nasledujici by vedlo k chybnemu reseni:
;; zachytil by se symbol `new-unnamed-symbol' a k reseni by to nevedlo
(define-macro no-capture
  (lambda body
    (let ((new-unnamed-symbol (gensym)))
      `(let ((new-unnamed-symbol 100))
	 ,@body))))

(let ((curval 10))
  (no-capture
   (display "Hodnota: ")
   (display new-unnamed-symbol)
   (newline)
   (+ curval 1)))

;;;;;;;;;;;;;;;;;;;; REALNY PRIKLAD

;; speci. forma or: provedena cistym zpusobem
(define-macro or
  (lambda args
    (if (null? args)
	#f
	(if (null? (cdr args))
	    (car args)
	    (let ((result (gensym)))
	      `(let ((,result ,(car args)))
		 (if ,result
		     ,result
		     (or ,@(cdr args)))))))))

; v poradku (jednonasobne vyhodnoceni):
(let ((x 0))
  (or (begin (set! x (+ x 1))
	     x)
      blah))		; ]=> 1

; taky v poradku (nedochazi k symbol capture):
(let ((result 10))
  (or #f result))	; ]=> 10

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; forma CASE
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; priklad pouziti:
(case (+ 1 2)
  ((0 1 2) 'blah)
  ((3 4) 'ahoj)
  (else 'nic))

;; naivni makro ``case''
(define-macro case
  (lambda (value . clist)
    `(let ((result ,value))
       (cond ,@(map (lambda (x)
		      (if (list? (car x))
			  `((member result ',(car x)) ,(cadr x))
			  `(else ,(cadr x))))
		    clist)))))

;; opet nefunguje:
(let ((result 1000))
  (case 10
    ((10 20) result)
    (else #f)))		; ]=> 10 ... misto 1000

;; reseni je zase jednoduche
(define-macro case
  (lambda (value . clist)
    (let ((result (gensym)))
      `(let ((,result ,value))
	 (cond ,@(map (lambda (x)
			(if (list? (car x))
			    `((member ,result ',(car x)) ,(cadr x))
			    `(else ,(cadr x))))
		      clist))))))

;; nyni jiz funguje jak ma:
(let ((result 1000))
  (case 10
    ((10 20) result)
    (else #f)))		; ]=> 1000

; ukazka pouziti
(case (* 2 3)
  ((2 3 5 7) 'prime)
  ((1 4 6 8 9) 'composite))

; ukazka pouziti
(case (car '(c d))
  ((a) 'a)
  ((b) 'b))
  
; ukazka pouziti
(case (car '(c d))
  ((a e i o u) 'vowel)
  ((w y) 'semivowel)
  (else 'consonant))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; full featured COND
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(cond)			; nedef.
(cond (else 'blah))	; blah
(cond ('blah))		; blah
(cond (10 => -))	; -10
(cond ((= 1 1)
       (display "X")
       (newline)
       (+ 1 2)))	; 3 (zobrazi X)

(define test
  (lambda (n)
    (cond ((= n 1) 'jedna)
	  ((= n 2))
	  ((= n 3) (display n)
                   (newline)
                   (+ n 1))
	  ((and (> n 4) n) =>
	   (lambda (x) (* x x)))
	  (else 'nevim))))

(test 0)	; nevim
(test 1)	; jedna
(test 2)	; #t
(test 3)	; 4 (zobrazi 3)
(test 10)	; 100

;; rozsireny cond (jako rekurzivni makro)
;; umoznuje mit specialni klauzuli tvaru => + vic vyrazu v tele
;; nebo prazdne telo, zde taky vyuzijeme zapamatovanou hodnotu
;; provedeno jako rekurzivni makro, s kazdou podminkou vlastni symbol
;; plus navic se vzdy vytvari nove prostredi
(define-macro cond
  (lambda clist
    (let ((symbol (gensym)))
      (if (null? clist)
	  `(if #f #f)
	  (if (equal? (caar clist) 'else)
	      `(begin ,@(cdar clist))
	      `(let ((,symbol ,(caar clist)))
		 (if ,symbol
		     ,(if (null? (cdar clist))
			  symbol
			  (if (equal? (cadar clist) '=>)
			      `(,(caddar clist) ,symbol)
			      `(begin ,@(cdar clist))))
		     (cond ,@(cdr clist)))))))))


;; pro kazdy cond je jeden generovany symbol
;; provedeno nejprve rozbalenim (mozna trochu slozitejsi)
(define-macro cond
  (lambda clist
    (let ((symbol (gensym)))
      `(let ((,symbol #f))
	 ,(let dive-ifs ((clist clist))
	    (if (null? clist)
		`(if #f #f)
		(if (equal? (caar clist) 'else)
		    `(begin ,@(cdar clist))
		    `(if (begin (set! ,symbol ,(caar clist))
				,symbol)
			 ,(if (null? (cdar clist))
			      symbol
			      (if (equal? (cadar clist) '=>)
				  `(,(caddar clist) ,symbol)
				  `(begin ,@(cdar clist))))
			 ,(dive-ifs (cdr clist))))))))))

; ukazka pouziti
(cond ((= 1 2) 'blah)
      ((= 1 3) 'ahoj 'svete)
      (#f => (lambda (x) (list x)))
      (#f)
      (else 1 2 3))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; system pro KVAZIKVOTOVANI (kdybychom nemeli)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; kvazikvotovani bez pouziti standardniho kvazikvotovani
(define-macro kvaziquote
  (lambda (expr)
    (if (not (list? expr))
	(list 'quote expr)
	(apply list 'apply 'append
	       (append
		(map (lambda (expr)
		       (if (and (list? expr)
				(not (null? expr)))
			   (cond ((equal? (car expr) 'unquote)
				  (list 'list (cadr expr)))
				 ((equal? (car expr) 'unquote-splicing)
				  (cadr expr))
				 ((equal? (car expr) 'kvaziquote)
				  (list 'list (list 'quote expr)))
				 (else (list 'list (list 'kvaziquote expr))))
			   (list 'list (list 'quote expr))))
		     expr)
		'((quote ())))))))

(kvaziquote 10)
(kvaziquote a)
(kvaziquote (+ 2 (unquote-splicing (reverse (quote (10 20 30)))) (* 3 4)))
(kvaziquote (+ 2 (unquote (reverse (quote (10 20 30)))) (* 3 4)))
(kvaziquote ((((1 '''(unquote (append '(1 2) '(3 4))) 3)))))
(kvaziquote (1 2 (kvaziquote (3 (unquote 4))) 5))
(kvaziquote (1 2 (unquote (kvaziquote (3 (unquote (+ 10 4))))) 5))
`(1 2 `(3 ,(+ 10 4)) 5)
`(1 2 ,`(3 ,(+ 10 4)) 5)
`(1 2 '(3 ,(+ 10 4)) 5)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; CYKLY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; cykly: do, while, repeat-until, for jako z C

;; cyklus typu ``WHILE'' (jako v jazyku C, vraci nedefinovanou hodnotu)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-macro while
  (lambda (condition . body)
    (let ((loop-name (gensym)))
      `(let ,loop-name ()
	    (if ,condition
		(begin ,@body
		       (,loop-name)))))))

(let ((i 0)
      (j 0))
  (while (< i 10)
	 (set! j (+ j i))
	 (set! i (+ i 1)))
  (list i j))	; (10 45)

;; uprava ``WHILE'' tak, aby vracel hodnotu vyhodnoceni posledniho vyrazu tela
(define-macro while
  (lambda (condition . body)
    (let ((loop-name (gensym))
	  (last-value (gensym)))
      `(let ,loop-name ((,last-value (if #f #f)))
	    (if ,condition
		(,loop-name (begin ,@body))
		,last-value)))))

(let ((i 0)
      (j 0))
  (while (< i 10)
	 (set! j (+ j i))
	 (set! i (+ i 1))
	 (list i j)))	; (10 45)

;; chyba DrScheme:
(while #f)	; konci chybou (pritom k tomu neni duvod)
(if #f (begin))	; DrScheme: begin: bad syntax (empty form) ...
(begin)		; pritom jede ... fakt divny

;; uprava: (begin #f ,@body)

;; cyklus typu ``FOR'' (C, PERL a dalsi)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-macro for
  (lambda (init condition incr . body)
    (let ((loop-name (gensym)))
      `(begin
	 ,init
	 (let ,loop-name ()
	      (if ,condition
		  (begin ,@body
			 ,incr 
			 (,loop-name))))))))

; priklad pouziti
(let ((i 0)
      (result 0))
  (for (set! i 5)
       (> i 0)
       (set! i (- i 1))
    (display (list "Stav: " i result))
    (newline)
    (set! result (+ result i)))
  (display (list "Koncovy: " i result))
  (newline))

;; CVICENI: pridejte mezi parametry FOR:
;; retval (vyraz, ktery se vyhodnoti po dokonceni behu cyklu)

;; cyklus typu ``DO'' (Scheme) pomoci ``letrec''
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-macro do
  (lambda (binding condition . body)
    (let ((loop-name (gensym)))
      `(letrec ((,loop-name
		 (lambda ,(map car binding)
		   (if ,(car condition)
		       (begin ,@(cdr condition))
		       (begin ,@body
			      (,loop-name ,@(map caddr binding)))))))
	 (,loop-name ,@(map cadr binding))))))

; priklad pouziti
(do ((x '(1 3 5 7 9) (cdr x))    ; navazany symbol
     (sum 0 (+ sum (car x))))    ; navazany symbol
    ((null? x) sum)              ; limitni podminka iterace
  (display (list x sum))         ; telo cyklu
  (newline))	; 25

;; cyklus typu ``DO'' pomoci pojmenovaneho letu (uprava predchoziho)
(define-macro do 
  (lambda (binding condition . body)
    (let ((loop-name (gensym)))
      `(let ,loop-name
	 ,(map (lambda (x)
		 (list (car x) (cadr x)))
	       binding)
	 (if ,(car condition)
	     (begin ,@(cdr condition))
	     (begin ,@body
		    (,loop-name ,@(map caddr binding))))))))

;; cyklus typu ``REPEAT-UNTIL'' (jako v jazyku Pascal)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; pomocna procedura: pro dany seznam vrat par (s . x), kde
;; `s' je vychozi seznam bez posledniho prvku a
;; `x' je posledni prvek vychoziho seznamu
(define but-last
  (lambda (l)
    (cond ((null? l) #f)
	  ((null? (cdr l)) (cons '() (car l)))
	  (else (let ((result (but-last (cdr l))))
		  (cons (cons (car l) (car result))
			(cdr result)))))))

; priklad pouziti:
(but-last '())		; #f
(but-last '(a b c d))	; ((a b c) . d)

;; nebo pro feinschmeckery (fajnsmekry), pomoci foldr:
(define but-last
  (lambda (l)
    (foldr (lambda (x y)
	     (if y
		 (cons (cons x (car y)) (cdr y))
		 (cons '() x)))
	   #f
	   l)))

;; makro realizujici cyklus typu `REPEAT-UNTIL'
(define-macro repeat
  (lambda args

    ;; interne definovany but-last
    ;; nektere interprety Scheme (treba DrScheme) vyhodnocuji transformacni
    ;; proceduru v separatnim prostredi, jehoz predkem je globalni prostredi
    ;; ale ve kterem nejsou zadne jine definice krome pocatecnich
    (define but-last
      (lambda (l)
	(cond ((null? l) #f)
	      ((null? (cdr l)) (cons '() (car l)))
	      (else (let ((result (but-last (cdr l))))
		      (cons (cons (car l) (car result))
			    (cdr result)))))))

    ;; zpracovani argumentu
    (let* ((split-args (but-last args))
	   (body (car split-args))
	   (limits (cdr split-args))
	   (loop-name (gensym)))

      ;; vlastni kod
      `(let ,loop-name ()
	    ,@body
	    (cond ,@(map (lambda (conds)
			   `(,(car conds) (begin ,@(cdr conds))))
			 (cdr limits))
		  (else (,loop-name)))))))

;; U PREDCHOZIHO BY STACILO DO CONDu DAT ,@(cdr limits) + (else (,loop- ...

; priklad pouziti
(let ((x 20)
      (y 15))
  (repeat
   (set! y (+ y 4))
   (set! x (- x 1))
   (until ((<= x 10) (list 'foo x y))
	  ((>= y 30) (list 'bar y (+ x 20))))))	; (bar 31 36)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; dalsi modifikace formy LET
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; TOHLE DAT AZ DO KNIZKY
;; elegantni reseni let-ref (let + refresh)

;; pomoci generovanych symbolu (DODELAT)

;; fluid-let: DAT GENEROVANE SYMBOLY
(define-macro fluid-let
  (lambda (xexe . body)
    (let ((xx (map car xexe))
          (ee (map cadr xexe))
          (old-xx (map (lambda (ig) (gensym)) xexe))
          (result (gensym)))
      `(let ,(map (lambda (old-x x) `(,old-x ,x)) 
                  old-xx xx)
         ,@(map (lambda (x e)
                  `(set! ,x ,e)) 
                xx ee)
         (let ((,result (begin ,@body)))
           ,@(map (lambda (x old-x)
                    `(set! ,x ,old-x)) 
                  xx old-xx)
           ,result)))))

;;  + dalsi verze letu (bez call/cc) -- na cviceni ???

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; HYGIENICKA MAKRA (kod prevzaty z R5RS)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; hygienicka makra: jiny pristup k vytvarenim maker
;;                   zamezuji vzniku `symbol capture',

;; zachovavaji lexikalni vazby, tim se mysli:
;; 1) Jestlize je v tele makra definovana vazba noveho symbolu,
;;    je automaticky prejmenovan tak, aby nekolidoval s zadnym jinym symbolem
;; 2) Pri vyhodnocovani tela makra se vazby volnych vyskytu symbolu
;;    hledaji v ramci lexikalne nadrazenych prostredi (to jest prostredi,
;;    ve kterych bylo makro definovano)

; vysvetlit: syntax-rules, let-syntax, letrec-syntax, define-syntax

; (define-syntax nazev transformacni-procedura)
; transformacni-procedura vznikla pouzitim spec. formy syntax-rules

; syntax-rules:
; vytvari transformacni proceduru pomoci zadanych vzoru pro prepis
; prvni argument je seznam literalu (klicova slova), muze byt prazdny
; dal nasleduji prepisovaci pravidla tvaru (vzor prepis)

; symboly vyskytujici se ve vzoru (krome prvniho) mohou byt bud:
; 1) symboly vyskytujici se mezi klicovymi slovy,
;    v tom pripade 
; 2) -||- nevyskystujici se mezi klicovymi slovy,
;    takove symboly jsou navazany na libovolny vstupni vyraz     
; pokud je za vzorem symbol ..., predpoklada se shoda (zadneho), jednoho,
; ci vice vstupnich elementu se vzorem:

;; `and' realizovane jako hygienicke makro
(define-syntax and
  (syntax-rules ()	    ; zadna klicova slova
    ((and) #t)              ; and bez argumentu
    ((and test) test)       ; and s jednim argumentem
    ((and test1 test2 ...)  ; and se dvema a vice argumenty
     (if test1 (and test2 ...) #f))))

;; `setf!' jako hygienicke makro
;; zde slouzi `car' `cdr' a `ref' jako klicova slova
;; pri pouziti makra musi byt uzivatelem pouzita
(define-syntax setf!
  (syntax-rules (car cdr ref)
    ((setf! (car pair) value) (set-car! pair value))
    ((setf! (cdr pair) value) (set-cdr! pair value))
    ((setf! (ref vector index) value) (vector-set! vector index value))
    ((setf! symbol value) (set! symbol value))))

; priklady pouziti:
(define x #f)
(setf! x 10)
x
(define p (cons 10 20))
(setf! (car p) 'ahoj)
p
(setf! (cdr p) 'svete)
p
(define l '(a b c d))
(setf! (car (cddr l)) 1)
l	; (a b 1 d)
(define v (vector 'a 'b 'c 'd))
(setf! (ref v 2) 10)
v	; #4(a b 10 d)

;; co by se stalo, kdybychom nezaradili `car', `cdr' a `ref' mezi literaly?
(define-syntax setf!
  (syntax-rules ()
    ((setf! (car pair) value) (set-car! pair value))
    ((setf! (cdr pair) value) (set-cdr! pair value))
    ((setf! (ref vector index) value) (vector-set! vector index value))
    ((setf! symbol value) (set! symbol value))))

; priklad proc nefunguje jak ma:
(define p (cons 10 20))
(setf! (cdr p) 'svete)	; pouzito je prvni pravidlo
p			; (svete . 20)

; duvod: car je pattern variable, ktera se navaze na symbol cdr
; tim padem vyrazu odpovida prvni pravidlo a provede se set-car!

;; `or' realizovane jako hygienicke makro
(define-syntax or
  (syntax-rules ()
    ((or) #f)
    ((or test) test)
    ((or test1 test2 ...)
     (let ((x test1))
       (if x x (or test2 ...))))))

; priklad:
(let ((x 10))
  (or #f x))	; 10 (spatna odpoved by byla #f)

;; `def' jako hygienicke makro
;; zde zalezi na poradi pravidel
(define-syntax def
  (syntax-rules ()
    ((def (name arg ...) stmt ...)
     (define name (lambda (arg ...) stmt ...)))
    ((def symbol stmt)
     (define symbol stmt))))

; prikaz `for' a la jazyk Pascal
(define-syntax for
  (syntax-rules (:= to downto do step)
    ((for var := start to end do stmt ...)
     (let loop ((var start))
       (if (<= var end)
	   (begin
	     stmt ...
	     (loop (+ var 1))))))
    ((for var := start downto end do stmt ...)
     (let loop ((var start))
       (if (>= var end)
	   (begin
	     stmt ...
	     (loop (- var 1))))))
    ((for var := start to end step inc do stmt ...)
     (let loop ((var start))
       (if (<= var end)
	   (begin
	     stmt ...
	     (loop (+ var inc))))))
    ((for var := start downto end step dec do stmt ...)
     (let loop ((var start))
       (if (>= var end)
	   (begin
	     stmt ...
	     (loop (- var dec))))))))

; priklady pouziti cyklu
(for i := 1 to 10 do (display i) (newline))
(for i := 10 downto 1 do (display i) (newline))
(for i := 1 to 10 step 2 do (display i) (newline))
(for i := 10 downto 1 step 2 do (display i) (newline))

;; JESTE K LEXIKALNIMU ROZSAHU
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; hygienicka makra je mozna definovat lokalne
;; pomoci forem let-syntax a letrec-syntax

;; priklad lokalni definice makra when v procedure
(define f
  (lambda (n)
    (let-syntax
	((when (syntax-rules ()
		 ((when test stmt1 ...)
		  (if test
		      (begin stmt1 ...))))))
      (when (> n 3)
	    (display "BLAH")
	    (newline)
	    (+ n 1)))))

; priklad pouziti:
(f 1)	; nedef.
(f 4)	; 5 (+ vytiskne BLAH)

(define f
  (lambda (n)
    (let-syntax
	((when (syntax-rules ()
		 ((when test stmt1 ...)
		  (if test
		      (begin stmt1 ...))))))
      (let ((test #f))
	(when (> n 3)
	      (display (list test "BLAH"))
	      (newline)
	      (+ n 1))))))

; priklad pouziti:
(f 1)	; nedef.
(f 4)	; 5 (+ vytiskne (#f BLAH))

;; letrec-syntax .. jako let-syntax, ale umoznuje, aby jednotliva
;; pravidla pouzivala ostatni pravidla, u let-syntax
;; se jednotliva pravidla ,,vzajemne nevidi''

;; tohle nebude fungovat, protoze & je definovane pomoci & (posledni pravidlo)
(let-syntax
    ((& (syntax-rules ()
	  ((&) #t)
	  ((& test) test)
	  ((& test1 test2 ...)
	   (if test1 (& test2 ...) #f)))))
  (& 1 2 3))	; error: & not bound

;; tohle uz bude fungovat
(letrec-syntax
    ((& (syntax-rules ()
	  ((&) #t)
	  ((& test) test)
	  ((& test1 test2 ...)
	   (if test1 (& test2 ...) #f)))))
  (& 1 2 3))	; 3
