
;;  SCHEME IN SCHEME
;;  Copyright (C) 2008  Vilem Vychodil, <vilem.vychodil@upol.cz>
;;
;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; version 2 of the License.
;;
;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.
;;
;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;;
;;;;  SYSTEM MANIFESTOVANYCH TYPU
;;;;

;; vytvor element jazyka s manifestovanym typem
(define curry-make-elem
  (lambda (type-tag)
    (lambda (data)
      (cons type-tag data))))

;; vrat visacku s typem
(define get-type-tag car)

;; vrat data
(define get-data cdr)

;; test daneho typu
(define curry-scm-type
  (lambda (type)
    (lambda (elem)
      (equal? type (get-type-tag elem)))))

;;;;
;;;;  ZAKLADNI ELEMENTY JAZYKA SCHEME
;;;;

;;;; CISLA
;;;;;;;;;;;;;;;;;;;;

;; konstruktor cisla a test zda-li se jedna o cislo
(define make-number (curry-make-elem 'number))
(define scm-number? (curry-scm-type 'number))

;;;; SYMBOLY
;;;;;;;;;;;;;;;;;;;;

;; konstruktor symbolu a test zda-li se jedna o symbol
(define make-symbol (curry-make-elem 'symbol))
(define scm-named-symbol? (curry-scm-type 'symbol))

;; konstruktor generovaneho symbolu
(define make-generated-symbol
  (let ((make-physical-element (curry-make-elem 'generated-symbol)))
    (lambda ()
      (make-physical-element (cons #f #f)))))
(define scm-generated-symbol? (curry-scm-type 'generated-symbol))

;; test zda-li se jedna o symbol pojmenovany/generovany
(define scm-symbol?
  (lambda (elem)
    (or (scm-named-symbol? elem)
	(scm-generated-symbol? elem))))

;;;; TECKOVE PARY 
;;;;;;;;;;;;;;;;;;;;

;; konstruktor paru CONS
(define make-pair
  (let ((make-physical-pair (curry-make-elem 'pair)))
    (lambda (head tail)
      (make-physical-pair (cons head tail)))))

;; test datoveho typu
(define scm-pair? (curry-scm-type 'pair))

;; selektor paru CAR
(define pair-car
  (lambda (pair)
    (if (scm-pair? pair)
	(car (get-data pair))
	(error "; Car: argument must be a pair"))))

;; selektor paru CDR
(define pair-cdr
  (lambda (pair)
    (if (scm-pair? pair)
	(cdr (get-data pair))
	(error "; Cdr: argument must be a pair"))))

;; mutator paru SET-CAR!
(define pair-set-car!
  (lambda (pair value)
    (if (scm-pair? pair)
	(begin
	  (set-car! (get-data pair) value)
	  the-void-value)
	(error "; Set-car!: argument must be a pair"))))

;; mutator paru SET-CDR!
(define pair-set-cdr!
  (lambda (pair value)
    (if (scm-pair? pair)
	(begin
	  (set-cdr! (get-data pair) value)
	  the-void-value)
	(error "; Set-cdr!: argument must be a pair"))))

;; obecny konvertor seznamu na scm-seznam a obracene
(define convert-list
  (lambda (a-null? a-car a-cdr b-cons b-nil f l)
    (if (a-null? l)
	b-nil
	(b-cons (f (a-car l))
		(convert-list a-null? a-car a-cdr b-cons b-nil f
			      (a-cdr l))))))

;;;; PROSTREDI
;;;;;;;;;;;;;;;;;;;;;;;;;

;; preved asociacni seznam na tabulku prostredi
(define assoc->env
  (lambda (l)
    (if (null? l)
	the-empty-list
	(make-pair (make-pair
		    (make-symbol (caar l))
		    (cdar l))
		   (assoc->env (cdr l))))))

;; konstruktor prostredi
(define make-env
  (let ((make-physical-env (curry-make-elem 'environment)))
    (lambda (pred table)
      (make-physical-env
       (cons pred (make-pair scm-false table))))))

;; test datoveho typu
(define scm-env? (curry-scm-type 'environment))

;; konstruktor globalniho prostredi
(define make-global-env
  (lambda (alist-table)
    (make-env scm-false (assoc->env alist-table))))

;; vrat tabulku
(define get-table
  (lambda (elem)
    (if (scm-env? elem)
	(cdr (get-data elem))
	(error "; Get-table: argument must be an environment"))))

;; vrat predka
(define get-pred
  (lambda (elem)
    (if (scm-env? elem)
	(car (get-data elem))
	(error "; Get-pred: argument must be an environment"))))

;; je globalni prostredi?
(define global?
  (lambda (elem)
    (and (scm-env? elem)
	 (equal? scm-false (get-pred elem)))))

;; hledani vazeb v asociacnim poli
(define scm-assoc
  (lambda (key alist)
    (cond ((scm-null? alist) scm-false)
	  ((eq? key (pair-car (pair-car alist))) (pair-car alist))
	  ((equal? key (pair-car (pair-car alist))) (pair-car alist))
	  (else (scm-assoc key (pair-cdr alist))))))

;; vyhledej vazbu v prostredi env, nebo vrat not-found
(define lookup-env
  (lambda (env symbol search-nonlocal? not-found)
    (let ((found (scm-assoc symbol (pair-cdr (get-table env)))))
      (cond ((not (equal? found scm-false)) found)
	    ((global? env) not-found)
	    ((not search-nonlocal?) not-found)
	    (else (lookup-env (get-pred env) symbol #t not-found))))))

;;;; PRIMITIVNI PROCEDURY
;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstruktor primitivni procedury a test zda-li se jedna o proceduru
(define make-primitive (curry-make-elem 'primitive))
(define scm-primitive? (curry-scm-type 'primitive))

;; vytvareni primitivnich procedur pomoci wrapperu
(define wrap-primitive
  (lambda (proc)
    (make-primitive
     (lambda arguments
       (expr->intern (apply proc (map get-data arguments)))))))

;; vytvareni predikatu pro testy typu pomoci wrapperu
(define wrap-predicate
  (lambda (type-pred)
    (make-primitive
     (lambda elems
       (if (apply type-pred elems)
	   scm-true
	   scm-false)))))

;;;; UZIVATELSKE PROCEDURY
;;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstruktor procedury
(define make-procedure
  (let ((make-physical-procedure (curry-make-elem 'procedure)))
    (lambda (env args body)
      (make-physical-procedure (list env args body)))))

;; selektory
(define procedure-environment (lambda (proc) (car (get-data proc))))
(define procedure-arguments (lambda (proc) (cadr (get-data proc))))
(define procedure-body (lambda (proc) (caddr (get-data proc))))

;; test zda-li se jedna o proceduru
(define scm-user-procedure? (curry-scm-type 'procedure))

;; test zda-li se jedna o primitivni/uzivatelskou proceduru
(define scm-procedure?
  (lambda (elem)
    (or (scm-primitive? elem)
	(scm-user-procedure? elem))))

;;;; PRIMITIVNI SPECIALNI FORMY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstruktor primitivni spec. formy a test zda-li se jedna o spec. formu
(define make-specform (curry-make-elem 'specform))
(define scm-specform? (curry-scm-type 'specform))

;; test zda-li se jedna o primitivni/uzivatelskou formu
(define scm-form?
  (lambda (elem)
    (or (scm-specform? elem)
	(scm-macro? elem))))

;;;; MAKRA
;;;;;;;;;;;;;;;;;;;;;;;;;;

;; konstrukto makra
(define make-macro (curry-make-elem 'macro))
(define scm-macro? (curry-scm-type 'macro))

;; test zda-li se jedna o primitivni/uzivatelskou formu
(define scm-form?
  (lambda (elem)
    (or (scm-specform? elem)
	(scm-macro? elem))))

;;;; SPECIALNI ELEMENTY JAZYKA
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; pravdivostni hodnoty
(define scm-false ((curry-make-elem 'boolean) #f))
(define scm-true  ((curry-make-elem 'boolean) #t))
(define scm-boolean? (curry-scm-type 'boolean))

;; prazdny seznam
(define the-empty-list ((curry-make-elem 'empty-list) '()))
(define scm-null? (lambda (elem) (eq? elem the-empty-list)))

;; vrazdna hodnota
(define the-void-value ((curry-make-elem 'void) '()))
(define scm-void? (lambda (elem) (equal? elem the-void-value)))

;; nedefinovana hodnota
(define the-undefined-value ((curry-make-elem 'undefined) '()))
(define scm-undefined? (lambda (elem) (equal? elem the-undefined-value)))

;;;; GENERICKE PROCEDURY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; scm-eq? naprogramujeme pomoci interniho eq?
(define scm-eq?
  (lambda (elem-a elem-b)
    (eq? (get-data elem-a)
	 (get-data elem-b))))

;;;;
;;;;  READER (nacteni vyrazu a prevedeni do interni representace)
;;;;

;; prevedeni vyrazu do interni formy
(define expr->intern
  (lambda (expr)
    (cond ((symbol? expr) (make-symbol expr))
	  ((number? expr) (make-number expr))
	  ((and (boolean? expr) expr) scm-true)
	  ((boolean? expr) scm-false)
	  ((null? expr) the-empty-list)
	  ((pair? expr) (make-pair (expr->intern (car expr))
				   (expr->intern (cdr expr))))
	  ((eof-object? expr) #f)
	  (else (error "; Syntactic error.")))))

;; nacti vstupni vyraz do interni formy
(define scm-read
  (lambda ()
    (expr->intern (read))))

;;;;
;;;;  PRINTER (vytisteni vyrazu v jeho externi reprez. -- pokud je to mozne)
;;;;

;; pouze pouzije display -- vypise syrovou reprezentaci
(define scm-print
  (lambda (elem)
    (display elem)))

;; prevedeni elementu do externi formy
(define intern->expr
  (lambda (expr)
    (cond ((scm-pair? expr) (cons (intern->expr (pair-car expr))
				  (intern->expr (pair-cdr expr))))
	  ((scm-env? expr) "#<environment>")
	  ((scm-primitive? expr) "#<primitive-procedure>")
	  ((scm-user-procedure? expr) "#<user-defined-procedure>")
	  ((scm-macro? expr) "#<macro>")
	  ((scm-generated-symbol? expr) "#<generated-symbol>")
	  ((scm-specform? expr) "#<special-form>")
	  ((scm-undefined? expr) "#<undefined>")
	  ((scm-void? expr) "#<void>")
	  (else (get-data expr)))))

;; o neco ,,chytrejsi printer''
(define scm-print
  (lambda (elem)
    (display (intern->expr elem))))

;;;;
;;;;  EVALUATOR
;;;;

;; map pres elementy tvorici scm-seznam, vysledkem je klasicky seznam
(define map-scm-list->list
  (lambda (f scm-l)
    (convert-list scm-null? pair-car pair-cdr cons '() f scm-l)))

;; preved scm-seznam na klasicky seznam
(define scm-list->list
  (lambda (scm-l)
    (convert-list scm-null? pair-car pair-cdr cons '() (lambda (x) x) scm-l)))

;; preved klasicky seznam na scm-seznam
(define list->scm-list
  (lambda (l)
    (convert-list null? car cdr make-pair the-empty-list (lambda (x) x) l)))

;; vyhodnot vyraz v danem prostredi
(define scm-eval
  (lambda (elem env)

    ;; vyhodnocovani elementu podle jejich typu
    (cond

     ;; symboly se vyhodnocuji na svou aktualni vazbu
     ((scm-symbol? elem)
      (let* ((binding (lookup-env env elem #t #f)))
	(if binding
	    (pair-cdr binding)
	    (error "; EVAL: Symbol not bound"))))

     ;; vyhodnoceni seznamu
     ((scm-pair? elem)

      ;; nejprve vyhodnotime prvni prvek seznamu
      (let* ((first (pair-car elem))
	     (args (pair-cdr elem))
	     (f (scm-eval first env)))

	;; podle prvniho prvku rozhodni o co se jedna
	(cond

	 ;; pokud se jedna o proceduru, vyhodnot argumenty a aplikuj
	 ((scm-procedure? f)
	  (scm-apply f (map-scm-list->list
			(lambda (elem)
			  (scm-eval elem env))
			args)))

	 ;; pokud se jedna o formu, aplikuj s nevyhodnocenymi argumenty
	 ((scm-form? f)
	  (scm-form-apply env f (scm-list->list args)))
	 
	 ;; na prvnim miste stoji nepripustny prvek
	 (error "; EVAL: First element did not eval. to procedure"))))

     ;; vse ostatni (cisla, boolean, ... se vyhodnocuje na sebe sama)
     (else elem))))

;; vytvor tabulku vazeb: formalni argument -- argument
(define make-bindings
  (lambda (formal-args args)
    (cond ((scm-null? formal-args) the-empty-list)
	  ((scm-symbol? formal-args)
	   (make-pair (make-pair formal-args (list->scm-list args))
		      the-empty-list))
	  (else (make-pair
		 (make-pair (pair-car formal-args) (car args))
		 (make-bindings (pair-cdr formal-args) (cdr args)))))))

;; aplikuj proceduru, predka nastav na ``env''
(define scm-env-apply
  (lambda (proc env args)
    (cond ((scm-primitive? proc) (apply (get-data proc) args))
	  ((scm-user-procedure? proc)
	   (scm-eval (procedure-body proc)
		     (make-env env
			       (make-bindings (procedure-arguments proc)
					      args))))
	  (else (error "; APPLY: Expected procedure")))))

;; aplikuj proceduru s lexikalnim predkem
(define scm-apply
  (lambda (proc args)
    (cond ((scm-primitive? proc) (scm-env-apply proc #f args))
	  ((scm-user-procedure? proc)
	   (scm-env-apply proc (procedure-environment proc) args))
	  (else (error "; APPLY: Expected procedure")))))

;; aplikuj spec. formu
(define scm-form-apply
  (lambda (env form args)
    (cond ((scm-specform? form) (apply (get-data form) env args))
	  ((scm-macro? form) (scm-eval (scm-apply (get-data form) args) env))
	  (else (error "; APPLY: Expected special form")))))

;;;;
;;;;  POMOCNE PROCEDURY
;;;;

;; pro obecny typ volani apply sestavi seznam argumentu
(define apply-collect-arguments
  (lambda (args)
    (cond ((null? args) (error "; APPLY: argument missing"))
	  ((and (not (null? args)) (null? (cdr args)))
	   (scm-list->list (car args)))
	  (else (cons (car args) (apply-collect-arguments (cdr args)))))))

;;;;
;;;;  TOPLEVEL ENVIRONMENT
;;;;

;; vytvor prostredi, ktere je nejvys v hierarchii
(define scheme-toplevel-env
  (make-global-env
   `(
     ;; specialni forma IF
     (if . ,(make-specform
	     (lambda (env condition expr . alt-expr)
	       (let ((result (scm-eval condition env)))
		 (if (equal? result scm-false)
		     (if (null? alt-expr)
			 the-undefined-value
			 (scm-eval (car alt-expr) env))
		     (scm-eval expr env))))))

     ;; specialni forma LAMBDA
     (lambda . ,(make-specform
		 (lambda (env args . body)
		   (if (not (null? (cdr body)))
		       (make-procedure
			env args
			(make-pair (make-symbol 'begin)
				   (let iter ((ar body))
				     (if (null? (cdr ar))
					 (make-pair (car ar) the-empty-list)
					 (make-pair (car ar)
						    (iter (cdr ar)))))))
		       (make-procedure env args (car body))))))

     ;; specialni forma QUOTE
     (quote . ,(make-specform (lambda (env elem) elem)))

     ;; specialni forma BEGIN
     (begin . ,(make-specform
		(lambda (env . body)
		  (let iter ((body body))
		    (cond ((null? body) the-undefined-value)
			  ((null? (cdr body)) (scm-eval (car body) env))
			  (else (begin 
				  (scm-eval (car body) env)
				  (iter (cdr body)))))))))

     ;; specialni forma DEFINE
     (define . ,(make-specform
		 (lambda (env symbol value)
		   (let ((value (scm-eval value env))
			 (result (lookup-env env symbol #f #f)))
		     (if result
			 (pair-set-cdr! result value)
			 (pair-set-cdr!
			  (get-table env)
			  (make-pair (make-pair symbol value)
				     (pair-cdr (get-table env)))))))))
     
     ;; specialni forma SET!
     (set! . ,(make-specform
	       (lambda (env symbol value)
		 (pair-set-cdr! (lookup-env env symbol #t #f)
				(scm-eval value env)))))

     ;; konstanty
     (pi . ,(make-number (* 4 (atan 1))))

     ;; procedury pro praci s cisly (zakomentovane procedury jsou v R5RS, ale
     ;; nejsou implementovany ve vsech dostupnych interpretech Scheme)
     (* . ,(wrap-primitive *))
     (+ . ,(wrap-primitive +))
     (- . ,(wrap-primitive -))
     (/ . ,(wrap-primitive /))
     (< . ,(wrap-primitive <))
     (<= . ,(wrap-primitive <=))
     (= . ,(wrap-primitive =))
     (> . ,(wrap-primitive >))
     (>= . ,(wrap-primitive >=))
     (abs . ,(wrap-primitive abs))
     (acos . ,(wrap-primitive acos))
;     (angle . ,(wrap-primitive angle))
     (asin . ,(wrap-primitive asin))
     (atan . ,(wrap-primitive atan))
     (ceiling . ,(wrap-primitive ceiling))
     (cos . ,(wrap-primitive cos))
;     (denominator . ,(wrap-primitive denominator))
     (even? . ,(wrap-primitive even?))
     (exact->inexact . ,(wrap-primitive exact->inexact))
     (exact? . ,(wrap-primitive exact?))
     (exp . ,(wrap-primitive exp))
     (expt . ,(wrap-primitive expt))
     (floor . ,(wrap-primitive floor))
     (gcd . ,(wrap-primitive gcd))
;     (imag-part . ,(wrap-primitive imag-part))
     (inexact->exact . ,(wrap-primitive inexact->exact))
     (inexact? . ,(wrap-primitive inexact?))
     (lcm . ,(wrap-primitive lcm))
     (log . ,(wrap-primitive log))
;     (magnitude . ,(wrap-primitive magnitude))
;     (make-polar . ,(wrap-primitive make-polar))
;     (make-rectangular . ,(wrap-primitive make-rectangular))
     (max . ,(wrap-primitive max))
     (min . ,(wrap-primitive min))
     (modulo . ,(wrap-primitive modulo))
     (negative? . ,(wrap-primitive negative?))
;     (numerator . ,(wrap-primitive numerator))
     (odd? . ,(wrap-primitive odd?))
     (positive? . ,(wrap-primitive positive?))
     (quotient . ,(wrap-primitive quotient))
;     (rationalize . ,(wrap-primitive rationalize))
;     (real-part . ,(wrap-primitive real-part))
     (remainder . ,(wrap-primitive remainder))
     (round . ,(wrap-primitive round))
     (sin . ,(wrap-primitive sin))
     (sqrt . ,(wrap-primitive sqrt))
     (tan . ,(wrap-primitive tan))
     (truncate . ,(wrap-primitive truncate))
     (zero? . ,(wrap-primitive zero?))

     ;; prace s pary
     (cons . ,(make-primitive make-pair))
     (car . ,(make-primitive pair-car))
     (cdr . ,(make-primitive pair-cdr))
     (set-car! . ,(make-primitive pair-set-car!))
     (set-cdr! . ,(make-primitive pair-set-cdr!))

     ;; negace
     (not . ,(make-primitive
	      (lambda (elem)
		(if (equal? elem scm-false)
		    scm-true
		    scm-false))))

     ;; predikaty
     (boolean? . ,(wrap-predicate scm-boolean?))
     (form? . ,(wrap-predicate scm-form?))
     (null? . ,(wrap-predicate scm-null?))
     (number? . ,(wrap-predicate scm-number?))
     (pair? . ,(wrap-predicate scm-pair?))
     (primitive? . ,(wrap-predicate scm-primitive?))
     (procedure? . ,(wrap-predicate scm-procedure?))
     (specform? . ,(wrap-predicate scm-specform?))
     (symbol? . ,(wrap-predicate scm-symbol?))
     (undefined? . ,(wrap-predicate scm-undefined?))
     (user-procedure? . ,(wrap-predicate scm-user-procedure?))
     (void? . ,(wrap-predicate scm-void?))
     (eq? . ,(wrap-predicate scm-eq?))
     
     ;; sektoru uzivatelskych procedur
     (procedure-environment . ,(make-primitive procedure-environment))
     (procedure-arguments . ,(make-primitive procedure-arguments))
     (procedure-body . ,(make-primitive procedure-body))

     ;; procedura environment-parent (vrat lexikalniho predka daneho prostredi)
     (environment-parent . ,(make-primitive get-pred))

     ;; procedura environment->list (vrat tabulku s vazbami v danem prosredi)
     (environment->list . ,(make-primitive
			    (lambda (elem)
			      (if (equal? elem scm-false)
				  scm-false
				  (get-table elem)))))

     ;; procedura eval (dvou argumentu, druhy je vzdy povinny)
     (eval . ,(make-primitive
	       (lambda (elem env)
		 (scm-eval elem env))))

     ;; procedura apply
     (apply . ,(make-primitive
		(lambda (proc . rest)
		  (scm-apply proc (apply-collect-arguments rest)))))

     ;; procedura macro + vracec transformatoru
     (macro . ,(make-primitive make-macro))
     (macro-transformer . ,(make-primitive get-data))

     ;; generovane symboly
     (gensym . ,(make-primitive make-generated-symbol))

     ;; procedura apply, ktere lze predat prostredi predka
     (env-apply . ,(make-primitive
		    (lambda (proc env . rest)
		      (scm-env-apply proc
				     env
				     (apply-collect-arguments rest)))))

     ;; procedura lookup-env
     (lookup-env . ,(make-primitive
		     (lambda (env symbol . nonlocal)
		       (lookup-env env symbol
				   (or (null? nonlocal)
				       (equal? (car nonlocal) scm-true))
				   scm-false))))

     )))

;; procedura load: vyhodnoceni vyrazu v souboru (slouzi k natazeni knihovny)
(define scm-load
  (lambda (file-name env)
    (let ((port (open-input-file file-name)))
      (let next ()
	(let ((expr (read port)))
	  (if (not (eof-object? expr))
	      (begin
		(scm-eval (expr->intern expr) env)
		(next)))))
      (close-input-port port))))

;; vrat nazev typu jako retezec s prvnim pismenem velkym
(define type-name
  (lambda (elem)
    (let ((strl (string->list (symbol->string (get-type-tag elem)))))
      (list->string (cons (char-upcase (car strl)) (cdr strl))))))

;;;;
;;;;  NOVY REPL
;;;;

;; cyklus REPL
(define scm-repl
  (lambda ()
    (let ((glob-env (make-env scheme-toplevel-env the-empty-list)))
      (scm-load "includes.scm" glob-env)
      (let loop ()
	(display "]=> ")
	(let ((elem (scm-read)))
	  (if (not elem)
	      'bye-bye
	      (let ((result (scm-eval elem glob-env)))
		(newline)
		(display (string-append "; " (type-name result)	": "))
		(scm-print result)
		(newline)
		(newline)
		(loop))))))))

;; spust REPL
(scm-repl)
